import asyncio
import importlib.metadata
from dataclasses import dataclass
from enum import Flag, auto
from pathlib import Path
from typing import Any, Callable, Coroutine, Iterable

from watchfiles import Change, awatch

from mountaineer.console import CONSOLE
from mountaineer.logging import LOGGER, pluralize
from mountaineer.paths import resolve_package_path


class CallbackType(Flag):
    CREATED = auto()
    MODIFIED = auto()
    DELETED = auto()


@dataclass
class CallbackEvent:
    action: CallbackType
    path: Path


@dataclass
class CallbackMetadata:
    # Since events can be debounced, we need to send all events that occurred
    # in the batch.
    events: list[CallbackEvent]


@dataclass
class CallbackDefinition:
    action: CallbackType
    callback: Callable[[CallbackMetadata], Coroutine[Any, Any, None]]


class FileWatcher:
    def __init__(
        self,
        callbacks: list[CallbackDefinition],
        ignore_list=["__pycache__", "_ssr", "_static", "_server", "_metadata"],
        ignore_hidden=True,
        debounce_interval=0.1,
    ):
        """
        :param debounce_interval: Seconds to wait for more events. Will only send one event per batched
        interval to avoid saturating clients with one action that results in many files.
        """
        self.callbacks = callbacks
        self.ignore_list = ignore_list
        self.ignore_hidden = ignore_hidden
        self.debounce_interval = debounce_interval
        self.debounce_task: asyncio.Task | None = None
        self.pending_events: list[CallbackEvent] = []

    def should_ignore_path(self, path: Path | str) -> bool:
        path_str = str(path)
        path_components = set(path_str.split("/"))

        # Check for any nested hidden directories and ignored directories
        if self.ignore_hidden and any(
            component.startswith(".") for component in path_components
        ):
            return True
        elif path_components & set(self.ignore_list) != set():
            return True
        return False

    def _map_change_to_callback_type(self, change: Change) -> CallbackType:
        if change == Change.added:
            return CallbackType.CREATED
        elif change == Change.modified:
            return CallbackType.MODIFIED
        elif change == Change.deleted:
            return CallbackType.DELETED
        # Default to modified for any other changes
        return CallbackType.MODIFIED

    async def _debounce(self, action: CallbackType, path: Path):
        if self.debounce_task is not None:
            self.debounce_task.cancel()

        self.pending_events.append(CallbackEvent(action=action, path=path))

        self.debounce_task = asyncio.create_task(self._handle_callbacks_after_delay())

    async def _handle_callbacks_after_delay(self):
        await asyncio.sleep(self.debounce_interval)
        await self.handle_callbacks()

    async def handle_callbacks(self):
        """
        Runs all callbacks for the given action.
        """
        for callback in self.callbacks:
            valid_events = [
                event
                for event in self.pending_events
                if event.action in callback.action
            ]
            if valid_events:
                await callback.callback(CallbackMetadata(events=valid_events))

        self.pending_events = []

    async def process_changes(self, changes: Iterable[tuple[Change, str]]):
        """
        Process a batch of changes from watchfiles.
        """
        for change, path_str in changes:
            path = Path(path_str)
            if self.should_ignore_path(path):
                continue

            action = self._map_change_to_callback_type(change)
            if not action.name:
                continue
            CONSOLE.print(f"[yellow]File {action.name.lower()}: {path}")
            await self._debounce(action, path)


class WatchdogLockError(Exception):
    def __init__(self, lock_path: Path):
        super().__init__(
            f"Watch lock file exists, another process may be running. If you're sure this is not the case, run:\n"
            f"`$ rm {lock_path}`",
        )
        self.lock_path = lock_path


class PackageWatchdog:
    """
    Monitors Python packages for file changes and triggers registered callbacks when changes occur.

    Manages the lifecycle of file watching across multiple packages, optimizing the watch paths
    to avoid redundant watchers, and verifying package availability. Used as the foundation for
    hot-reloading and development tooling in Mountaineer.

    ```python {{sticky: True}}
    import asyncio
    from mountaineer.development.watch import PackageWatchdog, CallbackDefinition, CallbackType, CallbackMetadata

    # Define a callback function to handle file changes
    async def reload_modules(metadata: CallbackMetadata) -> None:
        print(f"Changes detected in {len(metadata.events)} files")
        for event in metadata.events:
            print(f"  {event.action.name}: {event.path}")
            # You would typically reload modules or trigger other actions here

    # Create a watchdog for your main package and any dependencies
    watchdog = PackageWatchdog(
        main_package="my_app",
        dependent_packages=["my_library"],
        callbacks=[
            CallbackDefinition(
                action=CallbackType.MODIFIED | CallbackType.CREATED,
                callback=reload_modules
            )
        ],
        run_on_bootup=True
    )

    await watchdog.start_watching()
    ```

    """

    def __init__(
        self,
        main_package: str,
        dependent_packages: list[str],
        callbacks: list[CallbackDefinition] | None = None,
        run_on_bootup: bool = False,
    ):
        """
        Initialize a package watchdog to monitor file changes across multiple packages.

        :param main_package: Primary package to monitor for file changes
        :param dependent_packages: Additional packages to monitor for changes
        :param callbacks: List of callback definitions to execute when changes occur
        :param run_on_bootup: Typically, we will only notify callback if there has been
            a change to the filesystem. If this is set to True, we will run all callbacks
            on bootup as well.

        """
        self.main_package = main_package
        self.packages = [main_package] + dependent_packages
        self.paths: list[str] = []
        self.callbacks: list[CallbackDefinition] = callbacks or []
        self.run_on_bootup = run_on_bootup
        self.stop_event = asyncio.Event()
        self.running = False

    async def start_watching(self):
        """
        Begin asynchronously watching all package paths for file changes.

        If configured with run_on_bootup=True, immediately runs all callbacks once.
        Sets up the FileWatcher with registered callbacks and processes changes
        as they occur on the filesystem.

        :raises ValueError: If the watchdog is already running

        """
        self.check_packages_installed()
        self.get_package_paths()

        if self.run_on_bootup:
            for callback_definition in self.callbacks:
                await callback_definition.callback(CallbackMetadata(events=[]))

        if self.running:
            raise ValueError("Watchdog is already running")
        self.running = True

        watcher = FileWatcher(callbacks=self.callbacks)

        CONSOLE.print(
            f"👀 Watching {len(self.paths)} {pluralize(len(self.paths), 'path', 'paths')}"
        )
        for path in self.paths:
            LOGGER.info(f"Watching {path}")

        async for changes in awatch(
            *self.paths, stop_event=self.stop_event, watch_filter=None
        ):
            await watcher.process_changes(changes)

    def stop_watching(self):
        """
        Stop watching all file paths and reset the watchdog state.

        Signals the underlying watchfiles library to stop watching by setting
        the stop event and resets internal state for potential restart.
        """
        self.stop_event.set()
        self.stop_event = asyncio.Event()
        self.running = False

    def check_packages_installed(self):
        """
        Verify that all packages being watched are installed in the current environment.

        :raises ValueError: If any package is not installed
        """
        for package in self.packages:
            try:
                importlib.metadata.version(package)
            except importlib.metadata.PackageNotFoundError:
                raise ValueError(
                    f"Package '{package}' is not installed in the current environment"
                )

    def get_package_paths(self):
        """
        Resolve filesystem paths for all monitored packages.

        Finds each package's location on disk using importlib and optimizes
        the path list to eliminate redundant watchers through merge_paths.
        """
        paths: list[str] = []
        for package in self.packages:
            package_path = resolve_package_path(package)
            paths.append(str(package_path))
        self.paths = self.merge_paths(paths)

    def merge_paths(self, raw_paths: list[str]) -> list[str]:
        """
        Optimize the list of paths by removing subdirectories when their parent is already watched.

        If one path is a subdirectory of another, we only want to watch the parent
        directory. This function merges the paths to avoid duplicate watchers.

        :param raw_paths: List of directory paths to optimize

        :return List of optimized directory paths with redundancies removed
        """
        paths = [Path(path).resolve() for path in raw_paths]

        # Parents should come before their subdirectories
        paths.sort(key=lambda path: len(path.parts))

        merged: list[Path] = []

        for path in paths:
            # Check if the current path is a subdirectory of any path in the merged list
            if not any(path.is_relative_to(parent) for parent in merged):
                merged.append(path)

        # Convert Path objects back to strings
        return [str(path) for path in merged]



from mountaineer import LayoutControllerBase, RenderBase
from mountaineer.actions import sideeffect


class RootLayoutRender(RenderBase):
    layout_value: int
    layout_arg: int


class RootLayoutController(LayoutControllerBase):
    view_path = "/app/layout.tsx"

    def __init__(self):
        super().__init__()
        self.layout_value = 0

    def render(self, layout_arg: int | None = None) -> RootLayoutRender:
        return RootLayoutRender(
            layout_value=self.layout_value,
            layout_arg=layout_arg or 0,
        )

    @sideeffect
    async def increment_layout_value(self) -> None:
        self.layout_value += 1



from collections import defaultdict
from dataclasses import dataclass, field
from functools import wraps
from inspect import Parameter, Signature, isawaitable, isclass, signature
from json import JSONDecodeError, dumps as json_dumps, loads as json_loads
from pathlib import Path
from re import match as re_match
from time import monotonic_ns
from typing import Any, Callable, Optional, Type, cast, overload
from uuid import UUID, uuid4

from fastapi import APIRouter, FastAPI, Request
from fastapi.exceptions import RequestValidationError as RequestValidationErrorRaw
from fastapi.openapi.utils import get_openapi
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from inflection import underscore
from pydantic import BaseModel
from starlette.routing import BaseRoute

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.actions import (
    FunctionActionType,
    fuse_metadata_to_response_typehint,
    get_function_metadata,
    init_function_metadata,
)
from mountaineer.actions.fields import FunctionMetadata
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.client_compiler.base import APIBuilderBase
from mountaineer.client_compiler.build_metadata import BuildMetadata
from mountaineer.config import ConfigBase
from mountaineer.constants import DEFAULT_STATIC_DIR
from mountaineer.controller import ControllerBase
from mountaineer.controller_layout import LayoutControllerBase
from mountaineer.exceptions import (
    APIException,
    RequestValidationError,
    RequestValidationFailure,
)
from mountaineer.logging import LOGGER, debug_log_artifact
from mountaineer.paths import ManagedViewPath, resolve_package_path
from mountaineer.plugin import MountaineerPlugin
from mountaineer.render import Metadata, RenderBase, RenderNull
from mountaineer.ssr import find_tsconfig, render_ssr
from mountaineer.static import get_static_path


class ControllerDefinition(BaseModel):
    controller: ControllerBase
    router: APIRouter
    # URL prefix to the root of the server
    url_prefix: str
    # Dynamically generated function that actually renders the html content
    # This is a hybrid between render() and _generate_html()
    view_route: Callable
    # Render router is provided for all pages, only None for layouts that can't
    # be independently rendered as a webpage
    render_router: APIRouter | None

    model_config = {
        "arbitrary_types_allowed": True,
    }

    def get_url_for_metadata(self, metadata: FunctionMetadata):
        return f"{self.url_prefix}/{metadata.function_name.strip('/')}"


class ExceptionSchema(BaseModel):
    status_code: int
    schema_name: str
    schema_name_long: str
    schema_value: dict[str, Any]

    model_config = {
        "extra": "forbid",
    }


@dataclass
class LayoutElement:
    id: UUID

    # Can be null if a layout doesn't have a path
    controller: ControllerBase | None

    # Absolute path to the tsx / jsx entrypoint
    path: Path

    # Only models direct parent layouts at the next highest layer
    # Traverse these to find the full layout hierarchy
    parent: Optional["LayoutElement"] = None
    children: list["LayoutElement"] = field(default_factory=list)

    cached_server_script: str | None = None
    cached_server_sourcemap: str | None = None
    cached_client_script: str | None = None


class AppController:
    """
    Main entrypoint of a project web application.

    """

    builders: list[APIBuilderBase]

    global_metadata: Metadata | None
    """
    Metadata that's injected into every page. This is useful for setting
    global stylesheets and scripts. It's also useful for setting fallback
    metadata if it isn't overloaded by individual pages like setting a page title.

    """

    config: ConfigBase | None
    """
    The configuration object for the application. This is the main place
    to place runtime configuration values that might be changed across
    environments. One instance is registered as a global singleton and cached
    within the AppController class as well.

    """

    app: FastAPI
    """
    Internal FastAPI application instance used by Mountaineer. Exposed to add
    API-only endpoints and middleware.

    """

    fastapi_args: dict[str, Any] | None
    """
    Override or add additional arguments to the FastAPI constructor. See
    the FastAPI [documentation](https://fastapi.tiangolo.com/reference/fastapi/) for
    the attributes of this constructor.

    """

    live_reload_port: int
    """
    Port to use for live reloading of the webserver. This will be used to create a websocket connection
    to the browser to trigger a reload when the frontend has updates. By default
    we will bind to port 0 to guarantee an open system port.

    """

    controllers: list[ControllerDefinition]
    """
    Mounted controllers that are used to render the web application. Add
    a new one through `.register()`.

    """

    internal_api_prefix: str
    """
    URL prefix used for the action APIs that are auto-generated through @passthrough
    and @sideeffect decorators.

    """

    hierarchy_paths: dict[Path, LayoutElement]
    """
    Mapping of disk paths to file-system metadata about the controller. This internally
    builds up a DAG of the layout hierarchy to be used for rendering nested
    pages and layouts.

    """

    def __init__(
        self,
        *,
        name: str = "Mountaineer Webapp",
        version: str = "0.1.0",
        view_root: Path | None = None,
        global_metadata: Metadata | None = None,
        custom_builders: list[APIBuilderBase] | None = None,
        config: ConfigBase | None = None,
        fastapi_args: dict[str, Any] | None = None,
    ):
        self.app = FastAPI(title=name, version=version, **(fastapi_args or {}))
        self.controllers: list[ControllerDefinition] = []
        self.name = name
        self.version = version
        self.global_metadata = global_metadata
        self.builders = custom_builders if custom_builders else []

        self._controller_names: set[str] = set()

        # If this flag is present, we will re-raise this error during render()
        # so users can see the error in the browser.
        # This is useful for debugging, but should not be used in production
        self._build_exception: Exception | None = None

        # Follow our managed path conventions
        if config is not None and config.PACKAGE is not None:
            package_path = resolve_package_path(config.PACKAGE)
            self._view_root = ManagedViewPath.from_view_root(package_path / "views")
        elif view_root is not None:
            self._view_root = ManagedViewPath.from_view_root(view_root)
        else:
            raise ValueError(
                "You must provide either a config.package or a view_root to the AppController"
            )

        # Check our view directory is valid
        self._validate_view(self._view_root)

        # The act of instantiating the config should register it with the
        # global settings registry. We keep a reference to it so we can shortcut
        # to the user-defined settings later, but this is largely optional.
        self.config = config

        self.internal_api_prefix = "/internal/api"

        # The static directory has to exist before we try to mount it
        static_dir = self._view_root.get_managed_static_dir()

        # Mount the view_root / _static directory, since we'll need
        # this for the client mounted view files
        self.app.mount(
            DEFAULT_STATIC_DIR,
            StaticFiles(directory=str(static_dir)),
            name="static",
        )

        self.app.exception_handler(RequestValidationErrorRaw)(
            self._parse_validation_exception
        )
        self.app.exception_handler(APIException)(self._handle_exception)

        self.app.openapi = self.generate_openapi  # type: ignore

        # Edges that link the hierarchy together
        self.hierarchy_paths: dict[Path, LayoutElement] = {}

        self.live_reload_port: int = 0

    def _validate_view(self, view_root: ManagedViewPath):
        """
        Validates the view directory setup, including checking React version compatibility.

        :param view_root: The root directory containing the view files
        :raises ValueError: If any validation checks fail
        """
        package_json_path = view_root / "package.json"
        if not package_json_path.exists():
            LOGGER.warning(
                f"package.json not found at {package_json_path}. Please ensure your project has a valid package.json file."
            )
            return

        try:
            package_json = json_loads(package_json_path.read_text())
            react_version = package_json.get("dependencies", {}).get("react", "")
            if not react_version:
                react_version = package_json.get("devDependencies", {}).get("react", "")

            if not react_version:
                raise ValueError("React dependency not found in package.json")

            # Extract version number from semver string (e.g. "^19.0.0" -> "19.0.0")
            version_match = re_match(r"[\^~]?(\d+)\.\d+\.\d+", react_version)
            if not version_match:
                LOGGER.warning(f"Invalid React version format: {react_version}")
                return

            major_version = int(version_match.group(1))
            if major_version < 19:
                LOGGER.warning(
                    f"React version {react_version} is not supported. This application requires React 19.0 or higher."
                )
        except JSONDecodeError:
            LOGGER.warning(f"Invalid JSON in {package_json_path}")
        except Exception as e:
            LOGGER.warning(f"Error checking React version: {str(e)}")

    def register(self, controller: ControllerBase | MountaineerPlugin):
        """
        Register a new controller. This will:

        - Mount the html of the controller to the main application service
        - Mount all actions (ie. @sideeffect and @passthrough decorated functions) to their public API

        :param controller: The controller instance that should be added to your webapp. The class accepts a full
        instance instead of just a class, so you're able to perform any kind of runtime initialization of the
        kwarg args that you need before it's registered.

        """
        if isinstance(controller, ControllerBase):
            self._register_controller(controller)
        elif isinstance(controller, MountaineerPlugin):
            self._register_plugin(controller)
        else:
            raise ValueError(f"Unknown controller type: {type(controller)}")

    def _register_controller(self, controller: ControllerBase):
        # If we're running in production, sniff for the script files ahead of time and
        # attach them to the controller
        # This allows each view to avoid having to find these on disk, as well as gives
        # a proactive error if any view will be unable to render when their script files
        # are missing
        if not self.development_enabled:
            controller.resolve_paths(self._view_root, force=True)
            if not controller._bundled_scripts:
                raise ValueError(
                    f"Controller {controller} was not able to find its scripts on disk. Make sure to run your `build` CLI before starting your webapp."
                )
            if not controller._ssr_path:
                raise ValueError(
                    f"Controller {controller} was not able to find its server-side script on disk. Make sure to run your `build` CLI before starting your webapp."
                )

        self._register_controller_common(
            controller, dev_enabled=self.development_enabled
        )

    def _register_plugin(self, plugin: MountaineerPlugin):
        for controller in plugin.get_controllers():
            if isinstance(controller.view_path, str):
                controller.view_path = (
                    ManagedViewPath.from_view_root(plugin.view_root)
                    / controller.view_path
                )

            # This should find our precompiled static and ssr files
            controller._scripts_prefix = f"/static_plugins/{plugin.name}"
            controller.resolve_paths(plugin.view_root, force=True)

            if not controller._ssr_path or not controller._bundled_scripts:
                raise ValueError(
                    f"Controller {controller} was not able to find compiled scripts for plugin {plugin.name}"
                )

            # Dev mode is disabled so the app is forced to load the full built javascript
            # bundle when the pages load. This doesn't affect how the controller API endpoints
            # are mounted or otherwise how the view controller is added to the app.
            self._register_controller_common(controller, dev_enabled=False)

        # Mount the view_root / _static directory, since we'll need
        # this for the client mounted view files
        self.app.mount(
            f"/static_plugins/{plugin.name}",
            StaticFiles(directory=str(plugin.view_root / "_static")),
            name=f"static-{plugin.name}",
        )

    def _register_controller_common(
        self, controller: ControllerBase, dev_enabled: bool = True
    ):
        # Since the controller name is used to build dependent files, we ensure
        # that we only register one controller of a given name
        controller_name = controller.__class__.__name__
        if controller_name in self._controller_names:
            raise ValueError(
                f"Controller with name {controller_name} already registered."
            )

        # Update the paths now that we have access to the runtime package path
        controller_node = self.update_hierarchy(known_controller=controller)
        if not dev_enabled:
            if not controller._ssr_path:
                raise ValueError(
                    f"Controller {controller} was not able to find its server-side script on disk. Make sure to run your `build` CLI before starting your webapp."
                )
            controller_node.cached_server_script = controller._ssr_path.read_text()

        # The controller superclass needs to be initialized before it's
        # registered into the application
        if not hasattr(controller, "initialized"):
            raise ValueError(
                f"You must call super().__init__() on {controller} before it can be registered."
            )

        # We need to passthrough the API of the render function to the FastAPI router so it's called properly
        # with the dependency injection kwargs
        @wraps(controller.render)
        async def generate_controller_html(*args, **kwargs):
            start = monotonic_ns()

            # Figure out which controller we're rendering
            controller_node, direct_hierarchy = self._view_hierarchy_for_controller(
                controller,
            )

            direct_hierarchy.reverse()
            view_paths = [[str(layout.path) for layout in direct_hierarchy]]

            # Assemble the metadata for each controller involved in rendering this view
            # (this includes the current page and any wrapper LayoutControllers)
            render_overhead_by_controller = {}
            render_output = {}
            for node in direct_hierarchy:
                # If not set, must be a layout-only component. In this case we don't
                # need to do any rendering of internal data
                if not node.controller:
                    continue

                time = monotonic_ns()
                render_values = self._get_value_mask_for_signature(
                    signature(node.controller.render), kwargs
                )
                server_data = node.controller.render(**render_values)
                if isawaitable(server_data):
                    server_data = await server_data
                if server_data is None:
                    server_data = RenderNull()
                render_overhead_by_controller[node.controller.__class__.__name__] = (
                    monotonic_ns() - time
                )

                render_output[node.controller.__class__.__name__] = server_data

            # If the output of this controller's rendering is an explicit response, we should
            # just return that without any rendering
            controller_output = render_output[controller.__class__.__name__]
            if not isinstance(controller_output, RenderBase):
                return controller_output
            if (
                controller_output.metadata
                and controller_output.metadata.explicit_response
            ):
                return controller_output.metadata.explicit_response

            LOGGER.debug(
                f"Controller {controller.__class__.__name__} data acquired in {(monotonic_ns() - start) / 1e9}"
            )
            LOGGER.debug(
                f"Controller {controller.__class__.__name__} controller breakdown:\n"
                + "\n".join(
                    [
                        f"{controller_name}: {overhead / 1e9}"
                        for controller_name, overhead in render_overhead_by_controller.items()
                    ]
                )
            )

            # If we're in development mode, we should recompile the script on page
            # load to make sure we have the latest if there's any chance that it
            # was affected by recent code changes
            if dev_enabled:
                # Caching the build files saves about 0.3 on every load
                # during development
                start = monotonic_ns()

                # Find tsconfig.json in the parent directories of the view paths
                tsconfig_path = find_tsconfig(view_paths)

                if not controller_node.cached_server_script:
                    LOGGER.debug(
                        f"Compiling server-side bundle for {controller_name}: {view_paths}"
                    )
                    (
                        script_payloads,
                        sourcemap_payloads,
                    ) = mountaineer_rs.compile_independent_bundles(
                        view_paths,
                        str((self._view_root / "node_modules").resolve().absolute()),
                        "development",
                        0,
                        str(get_static_path("live_reload.ts").resolve().absolute()),
                        True,
                        tsconfig_path,
                    )
                    controller_node.cached_server_script = script_payloads[0]
                    controller_node.cached_server_sourcemap = sourcemap_payloads[0]
                if not controller_node.cached_client_script:
                    LOGGER.debug(
                        f"Compiling client-side bundle for {controller_name}: {view_paths}"
                    )
                    script_payloads, _ = mountaineer_rs.compile_independent_bundles(
                        view_paths,
                        str((self._view_root / "node_modules").resolve().absolute()),
                        "development",
                        self.live_reload_port,
                        str(get_static_path("live_reload.ts").resolve().absolute()),
                        False,
                        tsconfig_path,
                    )
                    controller_node.cached_client_script = script_payloads[0]
                LOGGER.debug(
                    f"Compiled dev scripts in {(monotonic_ns() - start) / 1e9}"
                )
                html = self.compile_html(
                    cast(str, controller_node.cached_server_script),
                    controller_output,
                    render_output,
                    inline_client_script=cast(
                        str, controller_node.cached_client_script
                    ),
                    external_client_imports=None,
                    sourcemap=controller_node.cached_server_sourcemap,
                )
            else:
                # Production payload
                html = self.compile_html(
                    cast(str, controller_node.cached_server_script),
                    controller_output,
                    render_output,
                    inline_client_script=None,
                    external_client_imports=[
                        f"{controller._scripts_prefix}/{script_name}"
                        for script_name in controller._bundled_scripts
                    ],
                    sourcemap=controller_node.cached_server_sourcemap,
                )

            LOGGER.debug(
                f"Controller {controller.__class__.__name__} load time took {(monotonic_ns() - start) / 1e9}"
            )
            return html

        # Strip the return annotations from the function, since we just intend to return an HTML page
        # and not a JSON response
        if not hasattr(generate_controller_html, "__wrapped__"):
            raise ValueError(
                "Unable to clear wrapped typehint, no wrapped function found."
            )

        return_model = generate_controller_html.__wrapped__.__annotations__.get(
            "return", MountaineerUnsetValue()
        )
        if isinstance(return_model, MountaineerUnsetValue):
            raise ValueError(
                "Controller render() function must have a return type annotation"
            )

        # Only the signature of the actual rendering function, not the original. We might
        # need to sniff render() again for its typehint
        generate_controller_html.__signature__ = signature(  # type: ignore
            generate_controller_html,
        ).replace(return_annotation=None)

        # Validate the return model is actually a RenderBase or explicitly marked up as None
        if not (
            return_model is None
            or (isclass(return_model) and issubclass(return_model, RenderBase))
        ):
            raise ValueError(
                "Controller render() return type annotation is not a RenderBase"
            )

        # Attach a new metadata wrapper to the original function so we can easily
        # recover it when attached to the class
        render_metadata = init_function_metadata(
            controller.render, FunctionActionType.RENDER
        )
        render_metadata.render_model = return_model

        # Register the rendering view to an isolated APIRoute, so we can keep track of its
        # the resulting router independently of the rest of the application
        # This is useful in cases where we need to do a render()->FastAPI lookup
        #
        # We only mount standard controllers, we don't expect LayoutControllers to have
        # a directly accessible URL
        if isinstance(controller, LayoutControllerBase):
            if hasattr(controller, "url"):
                raise ValueError(
                    f"LayoutControllers are not directly mountable to the router. {controller} should not have a url specified."
                )
            view_router = None
        else:
            view_router = APIRouter()
            view_router.get(controller.url)(generate_controller_html)
            self.app.include_router(view_router)
            render_metadata.register_controller_url(
                controller.__class__, controller.url
            )

        # Create a wrapper router for each controller to hold the side-effects
        controller_api = APIRouter()
        controller_url_prefix = (
            f"{self.internal_api_prefix}/{underscore(controller.__class__.__name__)}"
        )
        for _, fn, metadata in controller._get_client_functions():
            if not metadata.get_is_raw_response():
                # We need to delay adding the typehint for each function until we are here, adding the view. Since
                # decorators run before the class is actually mounted, they're isolated from the larger class/controller
                # context that the action function is being defined within. Here since we have a global view
                # of the controller (render function + actions) this becomes trivial
                return_model = fuse_metadata_to_response_typehint(
                    metadata, controller, render_metadata.get_render_model()
                )

                # Only mount the first time we register the function, otherwise we risk overwriting
                # the same function multiple times
                if controller.__class__ not in metadata.return_models:
                    metadata.register_return_model(controller.__class__, return_model)

                # Update the signature of the internal function, which fastapi will sniff for the return declaration
                # https://github.com/tiangolo/fastapi/blob/a235d93002b925b0d2d7aa650b7ab6d7bb4b24dd/fastapi/dependencies/utils.py#L207
                # Since these are consumed immediately by FastAPI, it's okay to overwrite previously set values (in the case
                # of superclass functions that are imported by multiple subclass controllers)
                method_function: Callable = fn.__func__  # type: ignore
                method_function.__signature__ = signature(method_function).replace(  # type: ignore
                    return_annotation=return_model
                )

            action_path = f"/{metadata.function_name}"
            controller_api.post(action_path)(fn)
            function_metadata = get_function_metadata(fn)
            function_metadata.register_controller_url(
                controller.__class__, f"{controller_url_prefix}{action_path}"
            )

        # Originally we tried implementing a sub-router for the internal API that was registered in the __init__
        # But the application greedily copies all contents from the router when it's added via `include_router`, so this
        # resulted in our endpoints not being seen even after calls to `.register(). We therefore attach the new
        # controller router directly to the application, since this will trigger a new copy of the routes.
        self.app.include_router(
            controller_api,
            prefix=controller_url_prefix,
        )

        LOGGER.debug(f"Did register controller: {controller_name}")

        controller_definition = ControllerDefinition(
            controller=controller,
            router=controller_api,
            view_route=generate_controller_html,
            url_prefix=controller_url_prefix,
            render_router=view_router,
        )
        controller._definition = controller_definition

        self.controllers.append(controller_definition)
        self._controller_names.add(controller_name)

        self._merge_hierarchy_signatures(controller_definition)

    def _view_hierarchy_for_controller(self, controller: ControllerBase):
        """
        Determines the nested parent layouts for the given controller, according
        to the currently mounted LayoutElement hierarchy.

        """
        # Figure out which controller we're rendering
        controller_node = next(
            node
            for node in self.hierarchy_paths.values()
            if node.controller == controller
        )
        # print("CONTROLLER", controller_node)

        # We need to figure out the layout controllers that should
        # wrap this controller
        direct_hierarchy: list[LayoutElement] = []
        current_node: LayoutElement | None = controller_node
        while current_node is not None:
            if current_node in direct_hierarchy:
                raise ValueError(f"Recursive layout detected: {current_node.path}")
            direct_hierarchy.append(current_node)
            current_node = current_node.parent

        return controller_node, direct_hierarchy

    def _merge_hierarchy_signatures(self, controller_definition: ControllerDefinition):
        # We should:
        # Update _this_ controller with anything in the above hierarchy (known layout controllers)
        # Update _child_ controller with this view (in the case that this definition is
        # a layout controller)
        node = next(
            node
            for node in self.hierarchy_paths.values()
            if node.controller == controller_definition.controller
        )

        def explore_children(node):
            for child in node.children:
                yield child
                yield from explore_children(child)

        def explore_parents(current_node):
            while current_node.parent is not None:
                yield current_node.parent
                current_node = current_node.parent

        parents = list(explore_parents(node))
        children = list(explore_children(node))

        parent_controllers = [node.controller for node in parents if node.controller]
        children_controllers = [node.controller for node in children if node.controller]

        parent_definitions = [
            controller_definition
            for controller_definition in self.controllers
            if controller_definition.controller in parent_controllers
        ]
        child_definitions = [
            controller_definition
            for controller_definition in self.controllers
            if controller_definition.controller in children_controllers
        ]

        for parent in parent_definitions:
            self.merge_render_signatures(
                controller_definition,
                reference_controller=parent,
            )
        for child in child_definitions:
            self.merge_render_signatures(
                child,
                reference_controller=controller_definition,
            )

    def invalidate_view(self, path: Path):
        """
        After an on-disk change of a given path, we should clear its current
        script cache so we rebuild with the latest changes. We should also clear
        out any nested children - so in the case of a layout change, we refresh
        all of its subpages.

        """
        if path.resolve().absolute() not in self.hierarchy_paths:
            # We have changed a path that isn't tracked as part of our
            # hierarchy. This is most likely a dependent file (like a component) that
            # is imported by some pages. Some early POC work has handled this explicitly
            # via parsing the whole project import dependences, but this introduces unnecessary
            # complexity when fresh-compile times are only ~0.3s and we only impact on dev.
            #
            # We allow components to clear all cached scripts and allow the next page
            # refresh to handle the rebuild.
            for path in list(self.hierarchy_paths):
                self.invalidate_view(path)

            return

        LOGGER.debug(f"Will invalidate path and children controllers: {path}")

        def _invalidate_node(node: LayoutElement):
            node.cached_server_script = None
            node.cached_client_script = None

            for child in node.children:
                _invalidate_node(child)

        node = self.hierarchy_paths[path]
        _invalidate_node(node)

    @overload
    def compile_html(
        self,
        server_script: str,
        page_metadata: RenderBase,
        all_render: dict[str, RenderBase],
        *,
        inline_client_script: str,
        external_client_imports: None,
        sourcemap: str | None,
    ): ...

    @overload
    def compile_html(
        self,
        server_script: str,
        page_metadata: RenderBase,
        all_render: dict[str, RenderBase],
        *,
        inline_client_script: None,
        external_client_imports: list[str],
        sourcemap: str | None,
    ): ...

    def compile_html(
        self,
        server_script: str,
        page_metadata: RenderBase,
        all_render: dict[str, RenderBase],
        *,
        inline_client_script: str | None = None,
        external_client_imports: list[str] | None = None,
        sourcemap: str | None = None,
    ):
        """
        Compiles the HTML for a given page, with all the controller-returned
        values hydrated into the page.

        """
        header_str: str
        if page_metadata.metadata:
            metadata = page_metadata.metadata
            if not metadata.ignore_global_metadata and self.global_metadata:
                metadata = metadata.merge(self.global_metadata)
            header_str = "\n".join(
                metadata.build_header(build_metadata=self.get_build_metadata())
            )
        else:
            if self.global_metadata:
                metadata = self.global_metadata
                header_str = "\n".join(
                    metadata.build_header(build_metadata=self.get_build_metadata())
                )
            else:
                header_str = ""

        # Client-side react scripts that will hydrate the server side contents on load
        server_data_json = {
            render_key: context.model_dump(mode="json")
            for render_key, context in all_render.items()
        }

        ssr_html = render_ssr(
            server_script,
            server_data_json,
            # TODO: Update to build param
            hard_timeout=10,
            sourcemap=sourcemap,
        )

        # Before building up the inline client script, log it to the temp directory
        # so we can inspect it in the debugger
        if inline_client_script is not None:
            debug_log_artifact("inline_client", "js", inline_client_script)

        client_import: str
        if inline_client_script:
            # We need to escape these inline. Otherwise we will close the parent script tag
            # prematurely and break the page.
            inline_client_script = inline_client_script.replace(
                "</script>", "<\\/script>"
            )

            # When we're running in debug mode, we just import
            # the script into each page so we can pick up on the latest changes

            # Wrap client-side code in an immediately invoked function expression (IIFE)
            # to isolate variables from the global scope and prevent conflicts with browser
            # built-in properties (e.g., 'chrome' in Chrome browser, 'safari' in Safari).
            # This prevents "duplicate variable" errors when client code defines variables
            # that match browser globals.
            client_import = f"""<script type='text/javascript'>
                (function() {{
                    {inline_client_script}
                }})();
                </script>"""
        elif external_client_imports:
            # This will point to our minified bundle that will in-turn import the other
            # common dependencies
            # Module types are required for browsers to use the import syntax
            # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Modules
            # All major browsers have had support since 2018
            client_import = "\n".join(
                [
                    f"<script type='module' src='{import_path}'></script>"
                    for import_path in external_client_imports
                ]
            )
        else:
            raise ValueError("Invalid client script import")

        page_contents = f"""
        <html>
        <head>
        {header_str}
        </head>
        <body>
        <div id="root">{ssr_html}</div>
        <script type="text/javascript">
        var SERVER_DATA = {json_dumps(server_data_json)};
        </script>
        {client_import}
        </body>
        </html>
        """

        return HTMLResponse(page_contents)

    def update_hierarchy(
        self,
        *,
        known_controller: ControllerBase | None = None,
        known_view_path: ManagedViewPath | None = None,
    ):
        """
        When we register a new element, we need to:

        - Find if there are any on-disk layouts that have been defined. We add these greedily, before
        we know if they're backed by a layout controller
        - Add the controller as a LayoutElement if it's new. Otherwise update the existing element

        This function is built to be invariant to the order of the updates, so we can call it multiple times
        and call it progressively as more controllers are added. It will be valid at
        any given state for the current webapp mounting.

        """
        if known_view_path is None:
            if known_controller is None:
                raise ValueError(
                    "Either known_controller or known_view_path must be provided"
                )

            full_view_path = (
                self._view_root / known_controller.view_path.lstrip("/")
                if isinstance(known_controller.view_path, str)
                else known_controller.view_path
            )
            full_view_path = full_view_path.resolve().absolute()
        else:
            full_view_path = known_view_path.resolve().absolute()

        # We should only update the current definition, we don't need to re-parse its hierarchy since
        # we assume the disk layout hasn't changed
        if full_view_path in self.hierarchy_paths:
            view_element = self.hierarchy_paths[full_view_path]
            if known_controller is not None:
                view_element.controller = known_controller
            return view_element

        view_element = LayoutElement(
            id=uuid4(), controller=known_controller, path=full_view_path
        )
        self.hierarchy_paths[full_view_path] = view_element

        # Recursively parse the parent paths to find the first layout (if any)
        # We go up until the view root. We allow the update_hierarchy to capture
        # the root view path, but we don't allow it to be a layout. This allows
        # layouts to inherit other layouts
        #
        # Resolve the path to the real underlying system path, necessary for /private/var
        # and /private symlinking in tmp paths
        current_path = full_view_path.realpath()
        package_root = full_view_path.get_root_link().realpath()
        while current_path != package_root:
            # We should never get to the OS root
            if str(current_path) == "/":
                raise ValueError(
                    f"View path ({full_view_path}) is not within the package root: {package_root}"
                )

            layout_file = current_path / "layout.tsx"
            if layout_file.exists():
                LOGGER.debug(
                    f"Layout found on disk, adding link: {view_element.path} {layout_file}"
                )
                parent_layout = self.update_hierarchy(known_view_path=layout_file)

                # Never create a self-referential layout
                # This can happen with layout controllers that will also find their
                # own layout.tsx file
                if parent_layout != view_element:
                    parent_layout.children.append(view_element)
                    view_element.parent = parent_layout

                    # We should break at the nearest level, since each page
                    # can only have one direct parent
                    break

            current_path = current_path.parent

        return view_element

    async def _handle_exception(self, request: Request, exc: APIException):
        return JSONResponse(
            status_code=exc.status_code,
            content=exc.internal_model.model_dump(),
        )

    async def _parse_validation_exception(
        self, request: Request, exc: RequestValidationErrorRaw
    ):
        raise RequestValidationError(
            errors=[
                RequestValidationFailure(
                    error_type=error["type"],
                    location=error["loc"],
                    message=error["msg"],
                    value_input=error["input"],
                )
                for error in exc.errors()
            ]
        )

    def merge_render_signatures(
        self,
        target_controller: ControllerDefinition,
        *,
        reference_controller: ControllerDefinition,
    ):
        """
        Collects the signature from the "reference_controller" and replaces the active target_controller's
        render endpoint with this new signature. We require all these new parameters to be kwargs so they
        can be injected into the render function by name alone.

        """
        reference_signature = signature(reference_controller.view_route)
        target_signature = signature(target_controller.view_route)

        reference_parameters = reference_signature.parameters.values()
        target_parameters = list(target_signature.parameters.values())

        # For any additional arguments provided by the reference, inject them into
        # the target controller
        # For duplicate ones, the target controller should win
        for parameter in reference_parameters:
            if parameter.name not in target_signature.parameters:
                target_parameters.append(
                    parameter.replace(
                        kind=Parameter.KEYWORD_ONLY,
                    )
                )
            else:
                # We only throw an error if the types are different. If they're the same we assume
                # that the resolution is intended to be shared.
                target_annotation_type = target_signature.parameters[
                    parameter.name
                ].annotation
                reference_annotation_type = parameter.annotation

                if target_annotation_type != reference_annotation_type:
                    raise TypeError(
                        f"Duplicate parameter {parameter.name} in {target_controller.controller} and {reference_controller.controller}.\n"
                        f"Conflicting types: {target_annotation_type} vs {reference_annotation_type}"
                    )

        target_controller.view_route.__signature__ = target_signature.replace(  # type: ignore
            parameters=target_parameters
        )

        # Re-mount the controller exactly as it was first mounted
        if target_controller.render_router:
            # Clear the previous definition before re-adding it
            # Both the app route is required (for the actual page resolution) and the render router
            # (to avoid conflicts in the OpenAPI generation)
            for route_list in [self.app.routes, target_controller.render_router.routes]:
                for route in list(route_list):
                    if (
                        isinstance(route, APIRoute)
                        and route.path == target_controller.controller.url
                        and route.methods == {"GET"}
                    ):
                        route_list.remove(route)

            target_controller.render_router.get(target_controller.controller.url)(
                target_controller.view_route
            )
            self.app.include_router(target_controller.render_router)

    def _get_value_mask_for_signature(
        self,
        signature: Signature,
        values: dict[str, Any],
    ):
        # Assume the values match the parameters specified in the signature
        passthrough_names = {
            parameter.name for parameter in signature.parameters.values()
        }
        return {
            name: value for name, value in values.items() if name in passthrough_names
        }

    def generate_openapi(self, routes: list[BaseRoute] | None = None):
        """
        Bundle custom user exceptions in the OpenAPI schema. By default
        endpoints just include the 422 Validation Error, but this allows
        for custom derived user methods.

        """
        openapi_base = get_openapi(
            title=self.name,
            version=self.version,
            routes=(routes if routes is not None else self.app.routes),
        )

        #
        # Exception injection
        #

        exceptions_by_url: dict[str, list[ExceptionSchema]] = {}
        for controller_definition in self.controllers:
            for (
                _,
                _,
                metadata,
            ) in controller_definition.controller._get_client_functions():
                url = controller_definition.get_url_for_metadata(metadata)
                # Not included in the specified routes, we should ignore this controller
                if url not in openapi_base["paths"]:
                    continue

                exceptions_models = metadata.get_exception_models()
                if not exceptions_models:
                    continue

                exceptions_by_url[url] = [
                    self._format_exception_model(exception_model)
                    for exception_model in exceptions_models
                ]

        # Users are allowed to reference the same schema name multiple times so long
        # as they have the same value. If they use conflicting values we'll have
        # to use the long name instead of the short module name to avoid conflicting
        # schema definitions.
        schema_names_to_long: defaultdict[str, set[str]] = defaultdict(set)
        for exception_payloads in exceptions_by_url.values():
            for payload in exception_payloads:
                schema_names_to_long[payload.schema_name].add(payload.schema_name_long)

        duplicate_schema_names = {
            schema_name
            for schema_name, schema_name_longs in schema_names_to_long.items()
            if len(schema_name_longs) > 1
        }

        for url, exception_payloads in exceptions_by_url.items():
            existing_status_codes: set[int] = set()

            for payload in exception_payloads:
                # Validate the exception state doesn't override existing values
                # Status codes are local to this particular endpoint but schema names
                # are global because they're placed in the global components section
                if payload.status_code in existing_status_codes:
                    raise ValueError(
                        f"Duplicate status code {payload.status_code} for {url}"
                    )

                schema_name = (
                    payload.schema_name
                    if payload.schema_name not in duplicate_schema_names
                    else payload.schema_name_long
                )

                other_definitions = {
                    definition_name: self._update_ref_path(definition)
                    for definition_name, definition in payload.schema_value.pop(
                        "$defs", {}
                    ).items()
                }
                openapi_base["components"]["schemas"].update(other_definitions)
                openapi_base["components"]["schemas"][schema_name] = (
                    self._update_ref_path(payload.schema_value)
                )

                # All actions are "posts" by definition
                openapi_base["paths"][url]["post"]["responses"][
                    str(payload.status_code)
                ] = {
                    "description": f"Custom Error: {payload.schema_name}",
                    "content": {
                        "application/json": {
                            "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                        }
                    },
                }

                existing_status_codes.add(payload.status_code)

        return openapi_base

    def _format_exception_model(self, model: Type[APIException]) -> ExceptionSchema:
        # By default all fields are optional. Since we are sending them
        # from the server we are guaranteed they will either be explicitly
        # provided or fallback to their defaults
        json_schema = model.InternalModel.model_json_schema()
        json_schema["required"] = list(json_schema["properties"].keys())

        return ExceptionSchema(
            status_code=model.status_code,
            schema_name=model.InternalModel.__name__,
            schema_name_long=f"{model.InternalModel.__module__}.{model.InternalModel.__name__}",
            schema_value=json_schema,
        )

    def _update_ref_path(self, schema: Any):
        """
        The $ref values that come out of the model schema are tied to #/defs instead
        of the #/components/schemas. This function updates the schema to use the
        correct prefix for the final OpenAPI schema.

        """
        if isinstance(schema, dict):
            new_schema: dict[str, Any] = {}
            for key, value in schema.items():
                if key == "$ref":
                    schema_name = value.split("/")[-1]
                    new_schema[key] = f"#/components/schemas/{schema_name}"
                    continue
                elif key == "additionalProperties":
                    # If the value is "False", we need to remove the key
                    if value is False:
                        continue

                new_schema[key] = self._update_ref_path(value)
            return new_schema
        elif isinstance(schema, list):
            return [self._update_ref_path(value) for value in schema]
        else:
            return schema

    def _definition_for_controller(
        self, controller: ControllerBase
    ) -> ControllerDefinition:
        for controller_definition in self.controllers:
            if controller_definition.controller == controller:
                return controller_definition
        raise ValueError(f"Controller {controller} not found")

    def get_build_metadata(self):
        """
        Will cache the build metadata in production but not in development, since
        we expect production developments will compile their metadata once and then
        use it for all endpoints.

        """
        if not self.development_enabled:
            # Determine if we've already cached the build
            if hasattr(self, "_build_metadata"):
                return getattr(self, "_build_metadata")

        metadata_path = self._view_root.get_managed_metadata_dir() / "metadata.json"
        if not metadata_path.exists():
            return None
        self._build_metadata = BuildMetadata.model_validate_json(
            metadata_path.read_text()
        )
        return self._build_metadata

    @property
    def development_enabled(self):
        return not self.config or self.config.ENVIRONMENT == "development"



from re import sub as re_sub
from typing import Any

import pytest

from mountaineer.client_builder.typescript import (
    TSLiteral,
    python_payload_to_typescript,
)


@pytest.mark.parametrize(
    "input_a,input_b,expected_literal",
    [
        (TSLiteral("a"), TSLiteral("b"), TSLiteral("ab")),
        (TSLiteral("a"), "b", TSLiteral("ab")),
        ("a", TSLiteral("b"), TSLiteral("ab")),
    ],
)
def test_tsliteral_combine(input_a: str, input_b: str, expected_literal: TSLiteral):
    result = input_a + input_b

    # We first do an explicit check of TSLiteral, making sure that we're not just
    # getting a regular string back
    assert isinstance(result, TSLiteral)
    assert result == expected_literal


@pytest.mark.parametrize(
    "payload, expected_str",
    [
        (1, "1"),
        (1.0, "1.0"),
        ("a", "'a'"),
        (True, "true"),
        (False, "false"),
        (None, "null"),
        (TSLiteral("a"), "a"),
        ({"a": "b"}, "{'a': 'b'}"),
        (["a", "b"], "['a', 'b']"),
    ],
)
def test_python_payload_to_typescript_primitives(payload: Any, expected_str: str):
    assert re_sub(r"\s+", "", python_payload_to_typescript(payload)) == re_sub(
        r"\s+", "", expected_str
    )


@pytest.mark.parametrize(
    "payload, expected_str",
    [
        (
            {"a": {"b": "b", "c": 1, "d": TSLiteral("someVariable")}},
            ("{\n  'a': {\n    'b': 'b',\n    'c': 1,\n    'd': someVariable\n  }\n}"),
        )
    ],
)
def test_python_payload_to_typescript_nested(payload: Any, expected_str: str):
    assert python_payload_to_typescript(payload) == expected_str


@pytest.mark.parametrize(
    "original_payload, expected_str",
    [
        # Should format as a consolidated literal
        (
            {"a": TSLiteral("a")},
            "{a}",
        ),
        # Not literal, should format as a string
        (
            {"a": "a"},
            "{'a': 'a'}",
        ),
        # Literal but different value, should map from a string to a variable
        (
            {"a": TSLiteral("b")},
            "{'a': b}",
        ),
    ],
)
def test_collapse_repeated_literals(
    original_payload: dict[str, str | TSLiteral], expected_str: str
):
    """
    If the key of our dictionary is a literal, and it's the same value as the key,
    we should collapse it into a single value.

    """
    assert re_sub(r"\s+", "", python_payload_to_typescript(original_payload)) == re_sub(
        r"\s+", "", expected_str
    )



import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum  # noqa
    from typing import Self  # noqa
else:
    """
    Full implementation of StrEnum, lifted from the Python 3.11 source code.

    https://github.com/python/cpython/blob/main/Lib/enum.py#L1337
    https://github.com/python/cpython/pull/22337
    https://github.com/python/cpython/pull/22362

    """
    from enum import Enum

    from typing_extensions import Self  # noqa

    class StrEnum(str, Enum):
        """
        Enum where members are also (and must be) strings
        """

        def __new__(cls, *values):
            "values must already be of type `str`"
            if len(values) > 3:
                raise TypeError("too many arguments for str(): %r" % (values,))
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):
                    raise TypeError("%r is not a string" % (values[0],))
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(values[1], str):
                    raise TypeError("encoding must be a string, not %r" % (values[1],))
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(values[2], str):
                    raise TypeError("errors must be a string, not %r" % (values[2]))
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        @staticmethod
        def _generate_next_value_(name, start, count, last_values):
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()

        __str__ = str.__str__  # type: ignore



from mountaineer import ConfigBase
from iceaxe.mountaineer import DatabaseConfig
from pydantic_settings import SettingsConfigDict

class AppConfig(ConfigBase, DatabaseConfig):
    PACKAGE: str | None = "{{project_name}}"

    model_config = SettingsConfigDict(env_file=(".env",))



class MountaineerUnsetValue:
    """
    A sentinel value to indicate that a value has not been set. This is useful for cases where
    we need to separate explicit Nones from missing values.
    """

    pass



import asyncio
import multiprocessing
from dataclasses import dataclass
from typing import Any

import pytest
import pytest_asyncio

from mountaineer.development.messages import IsolatedMessageBase, MessageTypes
from mountaineer.development.messages_broker import (
    AsyncMessageBroker,
    BrokerAuthenticationError,
    BrokerServerConfig,
    OKResponse,
)


@dataclass
class DummyMessage(IsolatedMessageBase):
    content: str

    def __eq__(self, other):
        # Allow equality checks in tests.
        return isinstance(other, DummyMessage) and self.content == other.content


@pytest_asyncio.fixture
async def broker_pair():
    """
    Provides a connected pair of brokers - a server broker and client broker.
    This simulates the typical usage pattern of the broker system.
    """
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        async with AsyncMessageBroker.new_client(config) as client_broker:
            yield server_broker, client_broker


@pytest.mark.asyncio
async def test_basic_job_flow(
    broker_pair: tuple[AsyncMessageBroker, AsyncMessageBroker],
):
    """Test the basic job flow between server and client brokers."""
    server_broker, client_broker = broker_pair

    # Send a job from server
    job_data = {"message": "test content"}
    job_id = "test-job-1"
    response = await server_broker.send_job(job_id, job_data)
    assert isinstance(response, OKResponse)

    # Client gets the job and sends response
    response_data = {"processed": True}
    await client_broker.send_response(job_id, response_data)

    # Server gets the response
    result = await server_broker.get_response(job_id)
    assert result == response_data


@pytest.mark.asyncio
async def test_auth_validation(
    broker_pair: tuple[AsyncMessageBroker, AsyncMessageBroker],
):
    """Test that authentication is properly enforced."""
    server_broker, _ = broker_pair

    # Create client with invalid auth
    invalid_config = BrokerServerConfig[MessageTypes](
        host=server_broker.host, port=server_broker.port, auth_key="invalid_key"
    )

    async with AsyncMessageBroker.new_client(invalid_config) as invalid_client:
        # Try to send a job with invalid auth
        with pytest.raises(BrokerAuthenticationError):
            await invalid_client.send_job("test-job", {"data": "test"})

        # Try to send a response with invalid auth
        with pytest.raises(BrokerAuthenticationError):
            await invalid_client.send_response("test-job", {"data": "test"})

        # Try to get a response with invalid auth
        with pytest.raises(BrokerAuthenticationError):
            await invalid_client.get_response("test-job")


@pytest.mark.asyncio
async def test_complex_object_serialization(
    broker_pair: tuple[AsyncMessageBroker, AsyncMessageBroker],
):
    """Test that complex Python objects can be sent through the broker."""
    server_broker, client_broker = broker_pair

    complex_data = {"numbers": [1, 2, 3], "nested": {"key": ["value1", "value2"]}}
    job_id = "complex-job"

    # Send complex job
    response = await server_broker.send_job(job_id, complex_data)
    assert isinstance(response, OKResponse)

    # Send complex response
    complex_response = {"processed": True, "data": [1, 2, 3]}
    await client_broker.send_response(job_id, complex_response)

    # Verify response
    result = await server_broker.get_response(job_id)
    assert result == complex_response


@pytest.mark.asyncio
async def test_multiple_jobs(
    broker_pair: tuple[AsyncMessageBroker, AsyncMessageBroker],
):
    """Test handling multiple jobs in sequence."""
    server_broker, client_broker = broker_pair

    # Send multiple jobs
    job_ids = [f"job-{i}" for i in range(3)]
    job_data = [{"message": f"message{i}"} for i in range(3)]

    # Send all jobs
    for job_id, data in zip(job_ids, job_data):
        response = await server_broker.send_job(job_id, data)
        assert isinstance(response, OKResponse)

    # Process all jobs
    for i, job_id in enumerate(job_ids):
        await client_broker.send_response(job_id, f"response{i}")

    # Verify all responses
    results = await asyncio.gather(
        *[server_broker.get_response(job_id) for job_id in job_ids]
    )
    assert results == ["response0", "response1", "response2"]


@pytest.mark.asyncio
async def test_nonexistent_job():
    """Test getting response for a nonexistent job."""
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        async with AsyncMessageBroker.new_client(config) as client_broker:
            # Try to get response for non-existent job
            # This should create a future that waits for the response
            get_response_task = asyncio.create_task(
                server_broker.get_response("nonexistent-job")
            )

            # Send response for the job
            await client_broker.send_response("nonexistent-job", "late response")

            # Now the get_response should complete
            result = await get_response_task
            assert result == "late response"


@pytest.mark.asyncio
async def test_concurrent_clients():
    """Test multiple clients can interact with the same server."""
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        async with (
            AsyncMessageBroker.new_client(config) as client1,
            AsyncMessageBroker.new_client(config) as client2,
        ):
            # Client 1 handles job1
            await server_broker.send_job("job1", "data1")
            await client1.send_response("job1", "response1")

            # Client 2 handles job2
            await server_broker.send_job("job2", "data2")
            await client2.send_response("job2", "response2")

            # Verify responses
            assert await server_broker.get_response("job1") == "response1"
            assert await server_broker.get_response("job2") == "response2"


async def run_client_process(
    config: BrokerServerConfig, job_id: str, response_data: Any
):
    """
    Async function to run inside a client process.
    """
    async with AsyncMessageBroker.new_client(config) as client:
        await client.send_response(job_id, response_data)


def client_process_entrypoint(config_dict: dict, job_id: str, response_data: Any):
    """
    Synchronous entry point for client process.
    """
    config = BrokerServerConfig[MessageTypes](**config_dict)
    asyncio.run(run_client_process(config, job_id, response_data))


@pytest.mark.asyncio
async def test_real_process_communication():
    """Test communication between real separate processes."""
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        # Convert config to dict for pickling
        config_dict = {
            "host": config.host,
            "port": config.port,
            "auth_key": config.auth_key,
        }

        # Create and start client process
        process = multiprocessing.Process(
            target=client_process_entrypoint,
            args=(config_dict, "test-job", "process-response"),
        )
        process.start()

        # Send job from server
        response = await server_broker.send_job("test-job", {"data": "test"})
        assert isinstance(response, OKResponse)

        # Wait for response from client process
        result = await server_broker.get_response("test-job")
        assert result == "process-response"

        # Clean up process
        process.join()
        assert process.exitcode == 0


@pytest.mark.asyncio
async def test_multiple_process_clients():
    """Test multiple client processes interacting with the server."""
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        config_dict = {
            "host": config.host,
            "port": config.port,
            "auth_key": config.auth_key,
        }

        # Create multiple client processes
        processes = []
        job_ids = [f"job-{i}" for i in range(3)]
        responses = [f"process-response-{i}" for i in range(3)]

        for job_id, response in zip(job_ids, responses):
            process = multiprocessing.Process(
                target=client_process_entrypoint, args=(config_dict, job_id, response)
            )
            process.start()
            processes.append(process)

        # Send jobs from server
        for job_id in job_ids:
            response = await server_broker.send_job(job_id, {"data": job_id})
            assert isinstance(response, OKResponse)

        # Get all responses
        results = await asyncio.gather(
            *[server_broker.get_response(job_id) for job_id in job_ids]
        )
        assert results == responses

        # Clean up processes
        for process in processes:
            process.join()
            assert process.exitcode == 0


@pytest.mark.asyncio
async def test_process_reconnection():
    """
    Test the edge case where:
    - A job is processed by one client process
    - That process exits
    - A new job is processed by a new client process
    """
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        config_dict = {
            "host": config.host,
            "port": config.port,
            "auth_key": config.auth_key,
        }

        # First client process
        process1 = multiprocessing.Process(
            target=client_process_entrypoint, args=(config_dict, "job1", "response1")
        )
        process1.start()

        # Send first job
        await server_broker.send_job("job1", "data1")
        result1 = await server_broker.get_response("job1")
        assert result1 == "response1"
        process1.join()
        assert process1.exitcode == 0

        # Second client process
        process2 = multiprocessing.Process(
            target=client_process_entrypoint, args=(config_dict, "job2", "response2")
        )
        process2.start()

        # Send second job
        await server_broker.send_job("job2", "data2")
        result2 = await server_broker.get_response("job2")
        assert result2 == "response2"
        process2.join()
        assert process2.exitcode == 0


def error_client_entrypoint(config_dict: dict):
    async def main():
        config = BrokerServerConfig[MessageTypes](**config_dict)
        async with AsyncMessageBroker.new_client(config) as client:
            # Simulate an error by trying to send response for non-existent job
            await client.send_response("nonexistent", "error")

    asyncio.run(main())


@pytest.mark.asyncio
async def test_process_error_handling():
    """Test handling of client process errors."""

    async with AsyncMessageBroker.start_server() as (server_broker, config):
        config_dict = {
            "host": config.host,
            "port": config.port,
            "auth_key": config.auth_key,
        }

        # Start error process
        process = multiprocessing.Process(
            target=error_client_entrypoint, args=(config_dict,)
        )
        process.start()
        process.join()

        # Process should exit cleanly even after error
        assert process.exitcode == 0


def worker_process(config_dict: dict):
    async def main():
        config = BrokerServerConfig[MessageTypes](**config_dict)
        async with AsyncMessageBroker.new_client(config) as client:
            # Get a job and send a response
            job_id, job_data = await client.get_job()
            await client.send_response(job_id, f"processed-{job_data['task']}")

    asyncio.run(main())


@pytest.mark.asyncio
async def test_get_job_process():
    """Test get_job functionality with real processes."""
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        config_dict = {
            "host": config.host,
            "port": config.port,
            "auth_key": config.auth_key,
        }

        # Start worker process
        process = multiprocessing.Process(target=worker_process, args=(config_dict,))
        process.start()

        # Send a job
        await server_broker.send_job("test-job", {"task": "test-task"})

        # Wait for response
        response = await server_broker.get_response("test-job")
        assert response == "processed-test-task"

        # Clean up process
        process.join()
        assert process.exitcode == 0


@pytest.mark.asyncio
async def test_multiple_waiting_clients():
    """Test multiple clients waiting for jobs."""
    async with AsyncMessageBroker.start_server() as (server_broker, config):
        async with (
            AsyncMessageBroker.new_client(config) as client1,
            AsyncMessageBroker.new_client(config) as client2,
        ):
            # Start both clients waiting for jobs
            task1 = asyncio.create_task(client1.get_job())
            task2 = asyncio.create_task(client2.get_job())

            # Small delay to ensure both clients are waiting
            await asyncio.sleep(0.1)

            # Send two jobs
            await server_broker.send_job("job1", {"task": "task1"})
            await server_broker.send_job("job2", {"task": "task2"})

            # Get results from both clients
            results = await asyncio.gather(task1, task2)

            # Verify each job was received exactly once
            job_ids = {result[0] for result in results}
            assert job_ids == {"job1", "job2"}



import asyncio
import gc
from contextlib import contextmanager
from random import random, seed
from typing import Any
from unittest.mock import MagicMock

import pytest

from mountaineer.cache import AsyncLoopObjectCache, extended_lru_cache

#
# LRU Cache
#


@extended_lru_cache(maxsize=100)
def get_random_data(param: int):
    return random()


@pytest.fixture(autouse=True)
def reset_cache():
    get_random_data._cache.clear()  # type: ignore


def test_extended_lru_cache():
    seed(20)
    val1 = get_random_data(0)
    val2 = get_random_data(0)

    assert val1 == val2


def test_extended_lru_cache_different_params():
    seed(20)
    val1 = get_random_data(0)
    val2 = get_random_data(1)

    assert val1 != val2


def test_extended_lru_no_cache():
    seed(20)
    val1 = get_random_data(0, use_cache=False)
    val2 = get_random_data(0, use_cache=False)

    assert val1 != val2


#
# AsyncLoopObjectCache
#


@contextmanager
def create_temporary_event_loop(set_running: bool = True):
    current_event_loop = asyncio._get_running_loop()

    # Out session cache requires the running loop to be set
    event_loop = asyncio.new_event_loop()
    asyncio._set_running_loop(event_loop if set_running else None)
    try:
        yield event_loop
    finally:
        event_loop.close()
        asyncio._set_running_loop(current_event_loop)


def test_get_obj_no_session():
    with create_temporary_event_loop():
        session_cache: AsyncLoopObjectCache[Any] = AsyncLoopObjectCache()
        assert session_cache.get_obj() is None


def test_set_and_get_obj():
    with create_temporary_event_loop():
        session_cache: AsyncLoopObjectCache[Any] = AsyncLoopObjectCache()
        mock_session = MagicMock()

        session_cache.set_obj(mock_session)
        obj = session_cache.get_obj()

        assert obj
        assert obj == mock_session


def test_get_lock_first_time():
    async def get_lock():
        session_cache: AsyncLoopObjectCache[Any] = AsyncLoopObjectCache()

        async with session_cache.get_lock():
            assert id(event_loop) in session_cache.loop_locks
            assert isinstance(session_cache.loop_locks[id(event_loop)], asyncio.Lock)

    with create_temporary_event_loop(set_running=False) as event_loop:
        event_loop.run_until_complete(get_lock())


def test_get_lock_reuse_lock():
    """
    get_lock should reuse an existing lock if one is already associated with the current
    event loop, so different tasks actually block on one another for the common object.

    """

    async def get_lock():
        session_cache: AsyncLoopObjectCache[Any] = AsyncLoopObjectCache()

        async with session_cache.get_lock():
            lock_first = session_cache.loop_locks[id(event_loop)]

        async with session_cache.get_lock():
            lock_second = session_cache.loop_locks[id(event_loop)]

        assert lock_first == lock_second

    with create_temporary_event_loop(set_running=False) as event_loop:
        event_loop.run_until_complete(get_lock())


def test_cleanup_on_loop_gc():
    async def set_and_get_obj(cache: AsyncLoopObjectCache[Any]):
        async with cache.get_lock():
            obj = "Temporary object"
            cache.set_obj(obj)
            assert cache.get_obj() == obj
            return cache

    # Create the first event loop and set an object in the cache
    with create_temporary_event_loop(set_running=False) as loop1:
        cache1 = AsyncLoopObjectCache[Any]()
        loop1.run_until_complete(set_and_get_obj(cache1))

    del loop1

    # Force garbage collection to cleanup the closed event loop
    gc.collect()

    # Verify that the object associated with the first loop has been cleaned up
    assert len(cache1.loop_caches) == 0
    assert len(cache1.loop_locks) == 0
    assert len(cache1.event_loop_refs) == 0

    # Create a second event loop and set another object in the cache
    with create_temporary_event_loop(set_running=False) as loop2:
        cache2 = AsyncLoopObjectCache[Any]()
        loop2.run_until_complete(set_and_get_obj(cache2))

    # Verify that the object is still present in the cache for the second loop
    assert len(cache2.loop_caches) == 1
    assert len(cache2.loop_locks) == 1
    assert len(cache2.event_loop_refs) == 1

    # Cleanup the second event loop
    del loop2
    gc.collect()

    # Verify that the object associated with the second loop has been cleaned up
    assert len(cache2.loop_caches) == 0
    assert len(cache2.loop_locks) == 0
    assert len(cache2.event_loop_refs) == 0



from create_mountaineer_app.environments.base import EnvironmentBase
from create_mountaineer_app.environments.poetry import PoetryEnvironment
from create_mountaineer_app.environments.uv import UvEnvironment
from create_mountaineer_app.environments.venv import VEnvEnvironment

__all__ = ["EnvironmentBase", "PoetryEnvironment", "UvEnvironment", "VEnvEnvironment"]



import warnings
from contextlib import AsyncExitStack, asynccontextmanager
from dataclasses import dataclass
from inspect import signature
from typing import Any, Callable

from fastapi import Request, params as fastapi_params
from fastapi.dependencies.utils import get_dependant, solve_dependencies


class DependenciesBaseMeta(type):
    """
    Dependencies have to be appended to their wrapper class explicitly. Providing static
    methods confuses the FastAPI resolution pipeline, because staticfunctions don't properly
    inspect as coroutines.

    Within `solve_dependencies`, it relies on function inspection to determine whether it should
    be run in the async loop or a separate thread. Executing `is_coroutine_callable` with a static
    method will always returns False, so we will inadvertantly run async dependencies in a thread
    loop. This will just return the raw coroutine instead of actually resolving the dependency.

    Adding functions to the class directly will just link their function signatures, which
    will inspect as intended.

    """

    def __new__(cls, name, bases, namespace, **kwargs):
        # Flag any child instances as deprecated but not the base model
        if name != "DependenciesBase":
            warnings.warn(
                (
                    "DependenciesBase is deprecated and will be removed in a future version.\n"
                    "Import modules to form dependencies. See mountaineer.dependencies.core for an example."
                ),
                DeprecationWarning,
                stacklevel=2,
            )

        for attr_name, attr_value in namespace.items():
            if isinstance(attr_value, staticmethod):
                raise TypeError(
                    f"Static methods are not allowed in dependency wrapper '{name}'. Found static method: '{attr_name}'."
                )
        return super().__new__(cls, name, bases, namespace, **kwargs)


class DependenciesBase(metaclass=DependenciesBaseMeta):
    pass


@dataclass
class DependencyOverrideProvider:
    # Internally FastAPI uses a property accessor to access the dependency_overrides, so we
    # reproduce this with a simple dataclass
    dependency_overrides: dict[Callable, Callable]


@asynccontextmanager
async def get_function_dependencies(
    *,
    callable: Callable,
    url: str | None = None,
    request: Request | None = None,
    dependency_overrides: dict[Callable, Callable] | None = None,
):
    """
    Get the dependencies of a function. This will return the values that should
    be injected into the function. Provide as much metadata as possible, so we can
    resolve more accurate dependencies. If not provided, we will synthesize some values.

    :param dependency_overrides: Specify functions that should be swapped-in when resolving
    the dependency chains. This is useful during testing or when you need to override one value
    in a dependency pipeline (like a user session) with a deterministic value.

    """
    # Synthesize defaults
    if not url:
        url = "/synthetic"
    if not request:
        request = Request(
            scope={
                "type": "http",
                "path": url,
                "path_params": {},
                "query_string": "",
                "headers": [],
            }
        )

    # Synthetic request object as if we're coming from the original first page
    dependant = get_dependant(
        call=callable,
        path=url,
    )

    async with AsyncExitStack() as async_exit_stack:
        payload = await solve_dependencies(
            request=request,
            dependant=dependant,
            async_exit_stack=async_exit_stack,
            dependency_overrides_provider=(
                DependencyOverrideProvider(dependency_overrides=dependency_overrides)
                if dependency_overrides
                else None
            ),
            embed_body_fields=False,
        )
        if payload.background_tasks:
            raise RuntimeError(
                "Background tasks are not supported when calling a static function, due to undesirable side-effects."
            )
        if payload.errors:
            raise RuntimeError(
                f"Errors encountered while resolving dependencies: {payload.errors}"
            )

        yield payload.values


def isolate_dependency_only_function(original_fn: Callable):
    """
    Create and return a mocked function that only includes the Depends parameters
    from the original function. This allows fastapi to resolve dependencies that are
    specified while allowing our logic to provide other non-dependency injected args.

    """
    sig = signature(original_fn)
    parameters = sig.parameters

    dependency_params = {
        name: param
        for name, param in parameters.items()
        if isinstance(param.default, fastapi_params.Depends)
    }

    # Construct a new function dynamically accepting only the dependencies
    async def mock_fn(**deps: Any) -> Any:
        pass

    mock_fn.__signature__ = sig.replace(parameters=dependency_params.values())  # type: ignore

    return mock_fn



from datetime import datetime

import pytest
from pydantic import BaseModel

from mountaineer.__tests__.client_builder.interface_builders.common import (
    create_exception_wrapper,
    create_field_wrapper,
    create_model_wrapper,
)
from mountaineer.actions.fields import FunctionActionType
from mountaineer.client_builder.interface_builders.action import ActionInterface
from mountaineer.client_builder.parser import (
    ActionWrapper,
    ExceptionWrapper,
    FieldWrapper,
    ModelWrapper,
)
from mountaineer.client_builder.types import ListOf, Or
from mountaineer.controller import ControllerBase
from mountaineer.exceptions import APIException


# Test Models
class StandardResponse(BaseModel):
    value: str
    timestamp: datetime


class ErrorResponse(APIException):
    error_code: str
    message: str


class FormData(BaseModel):
    name: str
    email: str


class AlternateResponse(BaseModel):
    status: str
    code: int


@pytest.fixture
def standard_response_wrapper():
    return create_model_wrapper(StandardResponse, "StandardResponse")


@pytest.fixture
def error_response_wrapper():
    return create_exception_wrapper(ErrorResponse, "ErrorResponse", 400)


@pytest.fixture
def form_data_wrapper():
    return create_model_wrapper(FormData, "FormData")


class TestBasicInterfaceGeneration:
    def test_simple_action_interface(self, standard_response_wrapper: ModelWrapper):
        action = ActionWrapper(
            name="simple_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=[],
            headers=[],
            request_body=None,
            response_bodies={ControllerBase: standard_response_wrapper},
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/base/simple_action"},
        )

        interface = ActionInterface.from_action(
            action, "/api/base/simple_action", ControllerBase
        )

        assert interface.name == "simple_action"
        assert "signal?: AbortSignal" in interface.typehints
        assert "Promise<StandardResponse>" in interface.response_type
        assert "StandardResponse" in interface.response_type

    @pytest.mark.parametrize(
        "params, expected_strs, expected_default_initializer",
        [
            # At least one required parameter, should require user input on functions
            (
                [
                    create_field_wrapper("required_param", str, True),
                    create_field_wrapper("optional_param", int, False),
                ],
                ["required_param: string", "optional_param?: number"],
                False,
            ),
            # All optional parameters, should not require user input on functions
            (
                [
                    create_field_wrapper("optional_param", int, False),
                ],
                ["optional_param?: number"],
                True,
            ),
        ],
    )
    def test_action_url_query_parameters(
        self,
        params: list[FieldWrapper],
        expected_strs: list[str],
        expected_default_initializer: bool,
    ):
        action = ActionWrapper(
            name="parametrized_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=params,
            headers=[],
            request_body=None,
            response_bodies={
                ControllerBase: create_model_wrapper(
                    StandardResponse, "StandardResponse"
                )
            },
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/main/parametrized_action"},
        )

        interface = ActionInterface.from_action(
            action, "/api/main/parametrized_action", ControllerBase
        )

        ts_code = interface.to_js()
        for expected_str in expected_strs:
            assert expected_str in ts_code
        assert interface.default_initializer == expected_default_initializer


class TestRequestBodyHandling:
    def test_form_action_interface(self, form_data_wrapper):
        action = ActionWrapper(
            name="form_action",
            module_name="test_module",
            action_type=FunctionActionType.SIDEEFFECT,
            params=[],
            headers=[],
            request_body=form_data_wrapper,
            response_bodies={
                ControllerBase: create_model_wrapper(
                    StandardResponse, "StandardResponse"
                )
            },
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/main/form"},
        )

        interface = ActionInterface.from_action(
            action, "/api/main/form", ControllerBase
        )

        ts_code = interface.to_js()
        assert "requestBody: FormData" in ts_code
        assert "mediaType" in "".join(interface.body)


class TestResponseTypeHandling:
    def test_raw_response_handling(self, standard_response_wrapper: ModelWrapper):
        action = ActionWrapper(
            name="raw_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=[],
            headers=[],
            request_body=None,
            response_bodies={ControllerBase: standard_response_wrapper},
            exceptions=[],
            is_raw_response=True,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/main/raw"},
        )

        interface = ActionInterface.from_action(action, "/api/main/raw", ControllerBase)

        assert "Promise<Response>" in interface.response_type

    def test_streaming_response_handling(self, standard_response_wrapper: ModelWrapper):
        action = ActionWrapper(
            name="stream_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=[],
            headers=[],
            request_body=None,
            response_bodies={ControllerBase: standard_response_wrapper},
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=True,
            controller_to_url={ControllerBase: "/api/main/stream"},
        )

        interface = ActionInterface.from_action(
            action, "/api/main/stream", ControllerBase
        )

        assert "AsyncGenerator<StandardResponse" in interface.response_type

    def test_multiple_response_types(self):
        class ResponseA(BaseModel):
            value: str

        class ResponseB(BaseModel):
            code: int

        class ControllerA(ControllerBase):
            pass

        class ControllerB(ControllerBase):
            pass

        action = ActionWrapper(
            name="multi_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=[],
            headers=[],
            request_body=None,
            response_bodies={
                ControllerA: create_model_wrapper(ResponseA, "ResponseA"),
                ControllerB: create_model_wrapper(ResponseB, "ResponseB"),
            },
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerA: "/api/a", ControllerB: "/api/b"},
        )

        interface = ActionInterface.from_action(action, "/api/multi", None)

        assert "ResponseA" in interface.response_type
        assert "ResponseB" in interface.response_type
        assert "|" in interface.response_type


class TestErrorHandling:
    def test_error_models(self, error_response_wrapper: ExceptionWrapper):
        action = ActionWrapper(
            name="error_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=[],
            headers=[],
            request_body=None,
            response_bodies={
                ControllerBase: create_model_wrapper(
                    StandardResponse, "StandardResponse"
                )
            },
            exceptions=[error_response_wrapper],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/main/error"},
        )

        interface = ActionInterface.from_action(
            action, "/api/main/error", ControllerBase
        )

        error_payload = "".join(interface.body)
        assert "ErrorResponse" in error_payload
        assert "errors" in error_payload
        assert "400" in error_payload


class TestTypeScriptGeneration:
    def test_complex_payload_conversion(self):
        action = ActionWrapper(
            name="complex_action",
            module_name="test_module",
            action_type=FunctionActionType.PASSTHROUGH,
            params=[
                create_field_wrapper("list_param", ListOf(str)),
                create_field_wrapper("optional_param", Or(str, None), False),
            ],
            headers=[],
            request_body=None,
            response_bodies={
                ControllerBase: create_model_wrapper(
                    StandardResponse, "StandardResponse"
                )
            },
            exceptions=[],
            is_raw_response=False,
            is_streaming_response=False,
            controller_to_url={ControllerBase: "/api/complex"},
        )

        interface = ActionInterface.from_action(action, "/api/complex", ControllerBase)

        ts_code = interface.to_js()
        assert "list_param: Array<string>" in ts_code
        assert "optional_param?: string" in ts_code



from enum import Enum, IntEnum

from pydantic import BaseModel, Field

from mountaineer.compat import StrEnum


class SubModel1(BaseModel):
    sub_a: str


class SubModel2(BaseModel):
    sub_b: int


class MyStrEnum(StrEnum):
    VALUE_1 = "value_1"
    VALUE_2 = "value_2"


class MyIntEnum(IntEnum):
    VALUE_1 = 1
    VALUE_2 = 2


class MyEnum(Enum):
    VALUE_1 = "value_1"
    VALUE_2 = 5


class MyModel(BaseModel):
    a: str = Field(description="The a field")
    b: int
    c: SubModel1
    d: list[SubModel1]
    both_sub: list[SubModel1 | SubModel2 | None]
    sub_map: dict[str, SubModel1 | None]



from pydantic import BaseModel


class BuildMetadata(BaseModel):
    """
    Metadata added during compile_time that should be maintained in the
    bundle for production hosting.

    """

    static_artifact_shas: dict[str, str]



import asyncio
from os import environ
from pathlib import Path
from subprocess import PIPE
from tempfile import TemporaryDirectory

from mountaineer.client_compiler.base import APIBuilderBase
from mountaineer.client_compiler.exceptions import BuildProcessException
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath


class PostCSSBundler(APIBuilderBase):
    """
    Support PostCSS processing for CSS files.

    - Assumes postcss-cli is installed in the primary project's package root (ie.
      the ci_webapp/views/node_modules directory)
    - Assumes that the css file under consideration has a root postcss.config.js file
      within its own view path

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # We want to keep track of known css files, so we'll specifically
        # rebuild them every change
        self.known_css_files: set[Path] = set()

    def mark_file_dirty(self, file_path: Path):
        # We expect that our runtime will call us at least once with the style
        if file_path.suffix in {".css", ".scss"}:
            self.known_css_files.add(file_path)
            super().mark_file_dirty(file_path)
        elif file_path.suffix in {".js", ".jsx", ".ts", ".tsx"}:
            # Stylesheets should reload if any frontend file or stylesheet changes,
            # not just the stylesheet itself since plugins like tailwind analyze the frontend
            # files for style declarations
            super().mark_file_dirty(file_path)

    async def build(self):
        if not self.metadata:
            raise ValueError("No metadata provided to build")

        if not self.dirty_files:
            LOGGER.debug("No dirty files, skipping postcss build")
            return

        known_css_files = self.managed_views_from_paths(list(self.known_css_files))

        # Figure out which roots the actually modified files belong to. We only
        # need to update the stylesheets that are in scope of the given root packages.
        dirty_managed = self.managed_views_from_paths(list(self.dirty_files))
        dirty_roots = {path.get_root_link() for path in dirty_managed}

        dirty_stylesheets = {
            stylesheet
            for stylesheet in known_css_files
            if stylesheet.get_root_link() in dirty_roots
        }
        LOGGER.debug(
            f"Potentially dirty stylesheets detected {dirty_stylesheets} of {known_css_files}"
        )

        # We only need to process the known css files
        for file_path in dirty_stylesheets:
            root_path = self.metadata.package_root_link
            built_css = await self.process_css(file_path)
            (
                root_path.get_managed_static_dir(tmp_build=True)
                / self.get_style_output_name(file_path)
            ).write_text(built_css)

    async def process_css(self, css_path: ManagedViewPath) -> str:
        """
        Process a CSS file using PostCSS and output the transformed contents.
        """
        if not self.metadata:
            raise ValueError("No metadata provided to build")

        is_installed, cli_path = self.postcss_is_installed(
            self.metadata.package_root_link
        )
        if not is_installed:
            raise EnvironmentError(
                f"postcss-cli is not installed in the specified view_root_path ({self.metadata.package_root_link}). Install it with:\n"
                "$ npm install -D postcss postcss-cli"
            )

        with TemporaryDirectory() as temp_dir_name:
            temp_dir_path = Path(temp_dir_name)
            output_path = temp_dir_path / "output.css"

            command = [
                str(cli_path.absolute()),
                str(css_path.absolute()),
                "-o",
                str(output_path.absolute()),
            ]
            LOGGER.debug(f"[postcss] RUNNING COMMAND {command}")

            process = await asyncio.create_subprocess_exec(
                *command,
                stdout=PIPE,
                stderr=PIPE,
                # postcss won't find the config file if we don't set the cwd
                # we assume this is in each view's root directory
                cwd=css_path.get_root_link().absolute(),
                env={
                    **environ,
                    # We expect the main package root will be the one with node modules installed
                    "NODE_PATH": str(
                        self.metadata.package_root_link.absolute() / "node_modules"
                    ),
                },
            )

            stdout, stderr = await process.communicate()

            if stdout.strip():
                LOGGER.info(stdout.decode())
            if stderr.strip():
                LOGGER.warning(stderr.decode())

            if process.returncode != 0:
                raise BuildProcessException(f"postcss error: {stderr.decode()}")

            return output_path.read_text()

    def get_style_output_name(self, original_stylesheet_path: ManagedViewPath) -> str:
        """
        Given a path to an original stylesheet, return the name of the compiled css file

        original_stylesheet_path: "path/to/styles.scss"
        output: "path_to_styles.css"
        """
        root_path = original_stylesheet_path.get_root_link()
        relative_path = original_stylesheet_path.relative_to(root_path)

        # Convert the parent portions of the path to make sure our final output
        # filename is unique
        unique_path = "_".join(
            [*relative_path.parent.parts, original_stylesheet_path.name]
        )
        return str(Path(unique_path).with_suffix(".css"))

    def postcss_is_installed(self, view_root_path: Path):
        # Adjust the check to look for local installation of postcss-cli, which we currently
        # require at the CLI bridge
        expected_path = view_root_path / "node_modules" / ".bin" / "postcss"
        return (
            expected_path.exists() and expected_path.is_file(),
            expected_path,
        )



from mountaineer.controller import ControllerBase


class LayoutControllerBase(ControllerBase):
    """
    Base class for layouts. Layout controllers are used to generate the HTML that wrap
    a regular view controller. They support all actions that a regular controller
    does (@sideeffect and @passthrough).

    Their limitations:
    - They are run in an isolated dependency injection context, they don't share
        dependency injected values with the given page
    - The current page Request is not supported within render()
    - Sideeffect updates to the layout don't affect the page state, and vice-versa
    - Layout controllers can't be mounted as a URL route

    """

    async def _generate_html(self, *args, global_metadata, **kwargs):
        raise NotImplementedError

    def _generate_ssr_html(self, server_data) -> str:
        raise NotImplementedError



# Re-export some core dependencies from other packages that will
# be used across projects
from fastapi import Depends as Depends

from mountaineer.actions.passthrough_dec import passthrough as passthrough
from mountaineer.actions.sideeffect_dec import sideeffect as sideeffect
from mountaineer.app import AppController as AppController
from mountaineer.client_compiler.postcss import PostCSSBundler as PostCSSBundler
from mountaineer.config import ConfigBase as ConfigBase
from mountaineer.controller import ControllerBase as ControllerBase
from mountaineer.controller_layout import LayoutControllerBase as LayoutControllerBase
from mountaineer.dependencies import CoreDependencies as CoreDependencies
from mountaineer.exceptions import APIException as APIException
from mountaineer.paths import ManagedViewPath as ManagedViewPath
from mountaineer.render import (
    LinkAttribute as LinkAttribute,
    MetaAttribute as MetaAttribute,
    Metadata as Metadata,
    RenderBase as RenderBase,
    ScriptAttribute as ScriptAttribute,
    ThemeColorMeta as ThemeColorMeta,
    ViewportMeta as ViewportMeta,
)



from enum import Enum
from pathlib import Path
from typing import List, Sequence

import pytest

from mountaineer.actions.passthrough_dec import passthrough
from mountaineer.actions.sideeffect_dec import sideeffect
from mountaineer.app import AppController
from mountaineer.client_builder.file_generators.base import ParsedController
from mountaineer.client_builder.file_generators.globals import (
    GlobalControllerGenerator,
    GlobalLinkGenerator,
)
from mountaineer.client_builder.parser import (
    ControllerParser,
    ControllerWrapper,
    EnumWrapper,
    ModelWrapper,
)
from mountaineer.controller import ControllerBase
from mountaineer.controller_layout import LayoutControllerBase
from mountaineer.paths import ManagedViewPath
from mountaineer.render import RenderBase


# Test Classes
class StatusEnum(Enum):
    ACTIVE = "active"
    PENDING = "pending"
    INACTIVE = "inactive"


class MainModel(RenderBase):
    name: str
    status: StatusEnum


class ChildModel(MainModel):
    child_field: str
    count: int


class DependentModel(MainModel):
    base: MainModel
    child: ChildModel
    current_status: StatusEnum


# Controllers
class BaseController(ControllerBase):
    @passthrough
    def base_action(self) -> MainModel:  # type: ignore
        pass


class ChildController(BaseController):
    url: str = "/child"
    view_path = "/child.tsx"

    async def render(self) -> DependentModel:  # type: ignore
        pass

    @sideeffect
    def update(self, data: ChildModel) -> DependentModel:  # type: ignore
        pass


class LayoutController(LayoutControllerBase):
    view_path = "/layout.tsx"

    async def render(self) -> MainModel:  # type: ignore
        pass


@pytest.fixture
def managed_path(tmp_path: Path) -> ManagedViewPath:
    return ManagedViewPath(tmp_path)


@pytest.fixture
def controller_parser() -> ControllerParser:
    return ControllerParser()


@pytest.fixture
def controller_wrappers(controller_parser: ControllerParser) -> list[ControllerWrapper]:
    # Concrete instances should be mounted to an AppController to augment
    # some of the runtime type information
    app_controller = AppController(view_root=Path())
    app_controller.register(ChildController())
    app_controller.register(LayoutController())

    return [
        controller_parser.parse_controller(ChildController),
        controller_parser.parse_controller(LayoutController),
    ]


# Tests
class TestGlobalControllerGenerator:
    @pytest.fixture
    def generator(
        self,
        managed_path: ManagedViewPath,
        controller_wrappers: List[ControllerWrapper],
    ) -> GlobalControllerGenerator:
        return GlobalControllerGenerator(
            managed_path=managed_path, controller_wrappers=controller_wrappers
        )

    def test_model_enum_graph_resolution(
        self, generator: GlobalControllerGenerator
    ) -> None:
        """Test that models and enums are sorted correctly"""
        # Get embedded types
        controllers = ControllerWrapper.get_all_embedded_controllers(
            generator.controller_wrappers
        )
        embedded = ControllerWrapper.get_all_embedded_types(
            controllers, include_superclasses=True
        )

        # Sort them
        sorted_items = generator._build_model_enum_graph(
            embedded.models, embedded.enums
        )

        # Hierarchy is:
        # StatusEnum
        # MainModel <- DependentModel
        #           <- ChildModel
        # Verify StatusEnum comes before BaseModel
        # Enums come before models and parents come before their subclasses
        enum_idx = self.get_item_order(sorted_items, "StatusEnum")
        main_model_idx = self.get_item_order(sorted_items, "MainModel")
        assert enum_idx < main_model_idx

        child_model_idx = self.get_item_order(sorted_items, "ChildModel")
        assert main_model_idx < child_model_idx

        dependent_model_idx = self.get_item_order(sorted_items, "DependentModel")
        assert main_model_idx < dependent_model_idx

    def test_controller_graph_resolution(
        self, generator: GlobalControllerGenerator
    ) -> None:
        """Test that controllers are sorted correctly"""
        controllers = ControllerWrapper.get_all_embedded_controllers(
            generator.controller_wrappers
        )
        sorted_controllers = generator._build_controller_graph(controllers)

        base_idx = self.get_item_order(sorted_controllers, "BaseController")
        child_idx = self.get_item_order(sorted_controllers, "ChildController")

        # Base should come before Child
        assert base_idx < child_idx

    def test_script_generation(self, generator: GlobalControllerGenerator) -> None:
        """Test the complete script generation"""
        blocks = generator.script()
        content = "\n".join(block.content for block in blocks)

        # Verify models are generated
        assert "export interface MainModel" in content
        assert "export interface ChildModel extends MainModel" in content
        assert "export interface DependentModel extends MainModel" in content

        # Verify enum is generated
        assert "export enum StatusEnum" in content
        assert "ACTIVE = " in content

        # Verify controllers are generated
        assert "export interface BaseController" in content
        assert "export interface ChildController extends BaseController" in content

        # Verify layout controller is included
        assert "export interface LayoutController" in content

    def get_item_order(
        self,
        sorted_items: Sequence[ModelWrapper | EnumWrapper | ControllerWrapper],
        raw_name: str,
    ):
        return next(
            i for i, item in enumerate(sorted_items) if item.name.raw_name == raw_name
        )


class TestGlobalLinkGenerator:
    @pytest.fixture
    def parsed_controllers(
        self, controller_parser: ControllerParser, managed_path: ManagedViewPath
    ) -> List[ParsedController]:
        (managed_path / "child").mkdir()
        (managed_path / "layout").mkdir()

        return [
            ParsedController(
                wrapper=controller_parser.parse_controller(ChildController),
                view_path=ManagedViewPath(managed_path / "child"),
                is_layout=False,
            ),
            ParsedController(
                wrapper=controller_parser.parse_controller(LayoutController),
                view_path=ManagedViewPath(managed_path / "layout"),
                is_layout=True,
            ),
        ]

    @pytest.fixture
    def generator(
        self, managed_path: ManagedViewPath, parsed_controllers: List[ParsedController]
    ) -> GlobalLinkGenerator:
        return GlobalLinkGenerator(
            managed_path=managed_path, parsed_controllers=parsed_controllers
        )

    def test_script_generation(self, generator: GlobalLinkGenerator) -> None:
        """Test link aggregator generation"""
        blocks = generator.script()
        content = "\n".join(block.content for block in blocks)

        # Verify imports
        assert "import { getLink as ChildControllerGetLinks }" in content

        # Verify layout controller is excluded
        assert "LayoutControllerGetLinks" not in content

        # Verify link generator object
        assert "const linkGenerator = {" in content
        assert "childController: ChildControllerGetLinks" in content
        assert "export default linkGenerator" in content



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

from mountaineer.controller import ControllerBase
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath


@dataclass
class ClientBundleMetadata:
    package_root_link: ManagedViewPath

    # We keep a tmpdir open for the duration of the build process, so our rust
    # logic can leverage file-based caches for faster builds
    # Note that this tmpdir is shared across all client builders, so it's important
    # that you enforce uniqueness of filenames if you leverage this cache
    tmp_dir: Path

    live_reload_port: int | None = None


class APIBuilderBase(ABC):
    """
    Base class for client builders. When mounted to an AppController, these build plugins
    will be called for every file defined in the view/app directory. It's up to the plugin
    whether to handle the incoming file.

    """

    def __init__(self):
        self.metadata: ClientBundleMetadata | None = None

        self.dirty_files: set[Path] = set()
        self.controllers: list[tuple[ControllerBase, ManagedViewPath]] = []

    def set_metadata(self, metadata: ClientBundleMetadata):
        self.metadata = metadata

    def register_controller(
        self, controller: ControllerBase, view_path: ManagedViewPath
    ):
        self.controllers.append((controller, view_path))

    def mark_file_dirty(self, file_path: Path):
        self.dirty_files.add(file_path)

    async def build_wrapper(self):
        """
        All internal users should use this instead of .build()
        """
        await self.build()
        self.dirty_files.clear()

    @abstractmethod
    async def build(self):
        """
        Builds the dirty files.

        """
        pass

    def managed_views_from_paths(self, paths: list[Path]) -> list[ManagedViewPath]:
        """
        Given a list of paths, assume these fall somewhere within the view directories
        specified by the controllers. Returns the ManagedViewPath objects for
        all paths where a match is found.

        """
        # Index all of the unique view roots to track the DAG hierarchies
        unique_roots = {view_path.get_root_link() for _, view_path in self.controllers}

        # Convert all of the dirty files into managed paths
        converted_paths: list[ManagedViewPath] = []
        for path in paths:
            # Each file must be relative to one of our known view roots, otherwise
            # we ignore it
            found_root = False
            for root in unique_roots:
                if path.is_relative_to(root):
                    relative_path = path.relative_to(root)
                    converted_paths.append(root / relative_path)
                    found_root = True
                    break

            if not found_root:
                LOGGER.warning(
                    f"File {path} is not relative to any view root ({unique_roots})"
                )

        return converted_paths



from enum import Enum
from typing import Generic, Optional, Type, TypeVar, cast

import pytest
from pydantic import BaseModel

from mountaineer.client_builder.aliases import AliasManager
from mountaineer.client_builder.parser import (
    ControllerParser,
    ControllerWrapper,
    EnumWrapper,
    ModelWrapper,
    SelfReference,
    WrapperName,
)
from mountaineer.controller import ControllerBase

T = TypeVar("T")


class TestAliasManager:
    @pytest.fixture
    def parser(self) -> ControllerParser:
        return ControllerParser()

    @pytest.fixture
    def alias_manager(self) -> AliasManager:
        return AliasManager()

    def test_basic_name_normalization(
        self, parser: ControllerParser, alias_manager: AliasManager
    ) -> None:
        class TestModel(BaseModel):
            field: str

        wrapper: ModelWrapper = ModelWrapper(
            name=WrapperName(TestModel.__name__),
            module_name="test.module",
            model=TestModel,
            isolated_model=TestModel,
            superclasses=[],
            value_models=[],
        )
        parser.parsed_models[TestModel] = wrapper

        alias_manager.assign_global_names(parser)
        assert wrapper.name.global_name == "TestModel"
        assert wrapper.name.raw_name == "TestModel"

    def test_global_model_conflict_resolution(
        self, parser: ControllerParser, alias_manager: AliasManager
    ) -> None:
        class User(BaseModel):
            name: str

        class User2(BaseModel):
            email: str

        User2.__name__ = "User"  # Force name conflict
        User2.__module__ = "auth.models"

        wrapper1: ModelWrapper = ModelWrapper(
            name=WrapperName(User.__name__),
            module_name="users.models",
            model=User,
            isolated_model=User,
            superclasses=[],
            value_models=[],
        )

        wrapper2: ModelWrapper = ModelWrapper(
            name=WrapperName(User2.__name__),
            module_name="auth.models",
            model=User2,
            isolated_model=User2,
            superclasses=[],
            value_models=[],
        )

        parser.parsed_models[User] = wrapper1
        parser.parsed_models[User2] = wrapper2

        alias_manager.assign_global_names(parser)

        assert wrapper1.name.global_name == "UsersModels_User"
        assert wrapper2.name.global_name == "AuthModels_User"

    def test_cross_type_conflict_resolution(
        self, parser: ControllerParser, alias_manager: AliasManager
    ) -> None:
        class Status1(BaseModel):
            code: int

        class Status2(Enum):
            ACTIVE = "active"
            INACTIVE = "inactive"

        Status1.__name__ = "Status"
        Status2.__name__ = "Status"

        model_wrapper: ModelWrapper = ModelWrapper(
            name=WrapperName("Status"),
            module_name="models.status",
            model=Status1,
            isolated_model=Status1,
            superclasses=[],
            value_models=[],
        )

        enum_wrapper: EnumWrapper = EnumWrapper(
            name=WrapperName("Status"), module_name="enums.status", enum=Status2
        )

        parser.parsed_models[Status1] = model_wrapper
        parser.parsed_enums[Status2] = enum_wrapper

        alias_manager.assign_global_names(parser)

        assert model_wrapper.name.global_name == "ModelsStatus_Status"
        assert enum_wrapper.name.global_name == "EnumsStatus_Status"

    def test_self_reference_updating(
        self, parser: ControllerParser, alias_manager: AliasManager
    ) -> None:
        class Node(BaseModel):
            value: str
            parent: Optional["Node"] = None

        class Node2(BaseModel):
            value: str
            parent: Optional[Node] = None

        wrapper: ModelWrapper = ModelWrapper(
            name=WrapperName("Node"),
            module_name="tree.models",
            model=Node,
            isolated_model=Node,
            superclasses=[],
            value_models=[],
        )

        # We just need to insert two of the same value so the duplicate detection
        # is triggered, their keys don't matter
        parser.parsed_models[Node] = wrapper
        parser.parsed_models[Node2] = wrapper
        parser.parsed_self_references.append(SelfReference(name="Node", model=Node))

        alias_manager.assign_global_names(parser)

        assert wrapper.name.global_name == "TreeModels_Node"
        assert parser.parsed_self_references[0].name == "TreeModels_Node"

    def test_local_name_resolution(
        self, parser: ControllerParser, alias_manager: AliasManager
    ) -> None:
        class StatusEnum(Enum):
            ACTIVE = "active"
            INACTIVE = "inactive"

        class UserModel(BaseModel):
            name: str
            status: StatusEnum

        class TestController(ControllerBase):
            url: str = "/test"

        controller_wrapper: ControllerWrapper = ControllerWrapper(
            name=WrapperName("TestController"),
            module_name="controllers",
            entrypoint_url="/test",
            controller=TestController,
            superclasses=[],
            queries=[],
            paths=[],
            render=None,
            actions={},
        )

        model_wrapper: ModelWrapper = ModelWrapper(
            name=WrapperName("UserModel"),
            module_name="models",
            model=UserModel,
            isolated_model=UserModel,
            superclasses=[],
            value_models=[],
        )

        enum_wrapper: EnumWrapper = EnumWrapper(
            name=WrapperName("StatusEnum"), module_name="enums", enum=StatusEnum
        )

        parser.parsed_controllers[TestController] = controller_wrapper
        parser.parsed_models[UserModel] = model_wrapper
        parser.parsed_enums[StatusEnum] = enum_wrapper

        alias_manager.assign_local_names(parser)

        assert model_wrapper.name.local_name == "UserModel"
        assert enum_wrapper.name.local_name == "StatusEnum"

    def test_generic_model_naming(
        self, parser: ControllerParser, alias_manager: AliasManager
    ) -> None:
        class Container(BaseModel, Generic[T]):
            value: T

        string_generic = cast(Type[BaseModel], Container[str])
        int_generic = cast(Type[BaseModel], Container[int])

        wrappers = {
            cls: ModelWrapper(
                name=WrapperName(cls.__name__),
                module_name=cls.__module__,
                model=cls,
                isolated_model=cls,
                superclasses=[],
                value_models=[],
            )
            for cls in [string_generic, int_generic]
        }

        for cls, wrapper in wrappers.items():
            parser.parsed_models[cls] = wrapper

        alias_manager.assign_global_names(parser)

        assert wrappers[string_generic].name.global_name == "ContainerStr"
        assert wrappers[int_generic].name.global_name == "ContainerInt"

    @pytest.mark.parametrize(
        "original_name, module, expected",
        [
            ("A", "test.module", "TestModule_A"),
            ("with_underscore", "test.module", "TestModule_WithUnderscore"),
            ("MyClass", "a.b.c.d", "ABCD_MyClass"),
            ("123Invalid", "test", "Test_123Invalid"),
            ("With Space", "test", "Test_WithSpace"),
        ],
    )
    def test_edge_cases(
        self,
        parser: ControllerParser,
        alias_manager: AliasManager,
        original_name: str,
        module: str,
        expected: str,
    ) -> None:
        class DynamicModel(BaseModel):
            field: str

        class OtherModel(BaseModel):
            field: str

        wrapper: ModelWrapper = ModelWrapper(
            name=WrapperName(original_name),
            module_name=module,
            model=DynamicModel,
            isolated_model=DynamicModel,
            superclasses=[],
            value_models=[],
        )

        parser.parsed_models[DynamicModel] = wrapper
        parser.parsed_models[OtherModel] = wrapper
        alias_manager.assign_global_names(parser)
        assert wrapper.name.global_name == expected

    @pytest.mark.parametrize(
        "module_path,expected_prefix",
        [
            ("users", "Users"),
            ("auth.models", "AuthModels"),
            ("api.v1.users.models", "ApiV1UsersModels"),
            ("my_app.user_models", "MyAppUserModels"),
            ("a.b.c", "ABC"),
            ("", ""),
            ("with.multiple.dots.at.end...", "WithMultipleDotsAtEnd"),
            ("___internal.models", "InternalModels"),
            ("api.v2", "ApiV2"),
            (
                "complex_name.with_underscores.and_numbers2",
                "ComplexNameWithUnderscoresAndNumbers2",
            ),
        ],
    )
    def test_typescript_prefix_from_module(
        self, alias_manager: AliasManager, module_path: str, expected_prefix: str
    ) -> None:
        """Test the module prefix formatting for various module path patterns"""
        result: str = alias_manager._typescript_prefix_from_module(module_path)
        assert result == expected_prefix



from uuid import uuid4

from locust import HttpUser, between, task


class SSRTest(HttpUser):
    """
    Load test the server SSR logic
    """

    wait_time = between(1, 2)

    def on_start(self):
        pass

    @task
    def simple_rendering(self):
        self.client.get("/")
        self.client.get(f"/detail/{uuid4()}/")

    @task
    def complex_rendering(self):
        self.client.get(f"/complex/{uuid4()}/")



from typing import AsyncIterator, Iterator, Optional

import pytest
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from mountaineer.actions.fields import (
    FunctionActionType,
    FunctionMetadata,
    ResponseModelType,
    annotation_is_metadata,
    extract_response_model_from_signature,
    fuse_metadata_to_response_typehint,
)
from mountaineer.controller import ControllerBase
from mountaineer.render import Metadata, RenderBase


class ExampleRenderModel(RenderBase):
    render_value_a: str
    render_value_b: int


class ExamplePassthroughModel(BaseModel):
    passthrough_value_a: str
    passthrough_value_b: int


class ExampleController(ControllerBase):
    async def render(self) -> None:
        pass


def basic_compare_model_fields(
    actual: dict[str, FieldInfo],
    expected: dict[str, FieldInfo],
):
    """
    FieldInfo objects can't be compared with native Python comparitors.
    """
    assert actual.keys() == expected.keys()
    for key in actual:
        assert actual[key].annotation == expected[key].annotation
        assert actual[key].default == expected[key].default
        assert actual[key].alias == expected[key].alias
        assert actual[key].is_required() == expected[key].is_required()


@pytest.mark.parametrize(
    "metadata, render_model, expected_model_name, expected_sideeffect_fields, expected_passthrough_fields",
    [
        # Case 1: Basic passthrough only
        (
            FunctionMetadata(
                action_type=FunctionActionType.PASSTHROUGH,
                function_name="example_function",
                passthrough_model=ExamplePassthroughModel,
            ),
            ExampleRenderModel,
            "ExampleFunctionResponseWrapped",
            None,
            {
                "passthrough_value_a": FieldInfo.from_annotation(annotation=str),
                "passthrough_value_b": FieldInfo.from_annotation(annotation=int),
            },
        ),
        # Case 2: Full sideeffect model (no reload states)
        (
            FunctionMetadata(
                action_type=FunctionActionType.SIDEEFFECT,
                function_name="example_function",
            ),
            ExampleRenderModel,
            "ExampleFunctionResponseWrapped",
            {
                "render_value_a": FieldInfo.from_annotation(annotation=str),
                "render_value_b": FieldInfo.from_annotation(annotation=int),
                "metadata": FieldInfo(annotation=Metadata | None, default=None),  # type: ignore
            },
            None,
        ),
        # Case 3: Sideeffect with reload states
        (
            FunctionMetadata(
                action_type=FunctionActionType.SIDEEFFECT,
                function_name="example_function",
                reload_states=tuple([ExampleRenderModel.render_value_a]),  # type: ignore
            ),
            ExampleRenderModel,
            "ExampleFunctionResponseWrapped",
            {
                "render_value_a": FieldInfo.from_annotation(annotation=str),
            },
            None,
        ),
        # Case 4: Combined passthrough and sideeffect
        (
            FunctionMetadata(
                action_type=FunctionActionType.SIDEEFFECT,
                function_name="example_function",
                passthrough_model=ExamplePassthroughModel,
                reload_states=tuple([ExampleRenderModel.render_value_a]),  # type: ignore
            ),
            ExampleRenderModel,
            "ExampleFunctionResponseWrapped",
            {
                "render_value_a": FieldInfo.from_annotation(annotation=str),
            },
            {
                "passthrough_value_a": FieldInfo.from_annotation(annotation=str),
                "passthrough_value_b": FieldInfo.from_annotation(annotation=int),
            },
        ),
    ],
)
def test_fuse_metadata_to_response_typehint(
    metadata: FunctionMetadata,
    render_model: type[RenderBase],
    expected_model_name: str,
    expected_sideeffect_fields: dict[str, FieldInfo] | None,
    expected_passthrough_fields: dict[str, FieldInfo] | None,
):
    sample_controller = ExampleController()
    result_model = fuse_metadata_to_response_typehint(
        metadata, sample_controller, render_model
    )

    # Verify model name
    assert result_model.__name__ == expected_model_name

    # Check sideeffect fields if expected
    if expected_sideeffect_fields:
        assert "sideeffect" in result_model.model_fields
        assert result_model.model_fields["sideeffect"].annotation
        basic_compare_model_fields(
            result_model.model_fields["sideeffect"].annotation.model_fields,
            expected_sideeffect_fields,
        )
    else:
        assert "sideeffect" not in result_model.model_fields

    # Check passthrough fields if expected
    if expected_passthrough_fields:
        assert "passthrough" in result_model.model_fields
        assert result_model.model_fields["passthrough"].annotation
        basic_compare_model_fields(
            result_model.model_fields["passthrough"].annotation.model_fields,
            expected_passthrough_fields,
        )
    else:
        assert "passthrough" not in result_model.model_fields


def test_fuse_metadata_inherited_render():
    """Test render inheritance with filtered fields"""

    class ParentRender(RenderBase):
        render_value_a: str
        render_value_b: int

    class ChildRender(ParentRender):
        render_value_c: float

    class TestController(ControllerBase):
        async def render(self) -> None:
            pass

    metadata = FunctionMetadata(
        action_type=FunctionActionType.SIDEEFFECT,
        function_name="test_function",
        reload_states=tuple([ParentRender.render_value_a]),  # type: ignore
    )

    result_model = fuse_metadata_to_response_typehint(
        metadata,
        TestController(),
        ChildRender,
    )

    # Should only include the specified parent field
    assert "sideeffect" in result_model.model_fields
    assert result_model.model_fields["sideeffect"].annotation

    model_fields = result_model.model_fields["sideeffect"].annotation.model_fields
    basic_compare_model_fields(
        model_fields,
        {
            "render_value_a": FieldInfo.from_annotation(annotation=str),
        },
    )


def test_fuse_metadata_to_response_typehint_unique_models():
    """
    Ensure that render_bases from different files will be considered unique
    in their (module, name) pairing. By default all created modules will inherit
    the mountaineer field module name, so they appear as duplicates to project-wide classes
    with the same individual name.

    """

    # Create render models with same name but different modules
    class RenderA(RenderBase):
        __module__ = "renders.a"
        name: str
        age: int

    class RenderB(RenderBase):
        __module__ = "renders.b"
        name: str
        address: str

    class MockControllerA(ControllerBase):
        __module__ = "controller_a"

        async def render(self) -> RenderA:
            return RenderA(name="example", age=1)

    class MockControllerB(ControllerBase):
        __module__ = "controller_b"

        async def render(self) -> RenderB:
            return RenderB(name="example", address="123 Fake St.")

    # Create two metadata instances with the same function name but different controllers
    metadata_a = FunctionMetadata(
        function_name="get_user", action_type=FunctionActionType.RENDER
    )
    metadata_b = FunctionMetadata(
        function_name="get_user", action_type=FunctionActionType.RENDER
    )

    # Generate response models for each controller/render pair
    response_a = fuse_metadata_to_response_typehint(
        metadata_a, MockControllerA(), RenderA
    )
    response_b = fuse_metadata_to_response_typehint(
        metadata_b, MockControllerB(), RenderB
    )

    # Verify the models have different modules and inherited the ones
    # from the render models
    assert response_a.__module__ == "controller_a"
    assert response_b.__module__ == "controller_b"


def test_annotation_is_metadata():
    assert annotation_is_metadata(Metadata)
    assert annotation_is_metadata(Optional[Metadata])  # type: ignore
    assert annotation_is_metadata(Metadata | None)  # type: ignore
    assert not annotation_is_metadata(str)


class ExampleModel(BaseModel):
    value: str


def test_extract_response_model_from_signature():
    def none_typehint() -> None:
        pass

    def example_typehint(self) -> ExampleModel:
        return ExampleModel(value="example")

    async def example_async_typehint(self) -> ExampleModel:
        return ExampleModel(value="example")

    def example_iterator_typehint(self) -> Iterator[ExampleModel]:
        yield ExampleModel(value="example")

    async def example_async_iterator_typehint(self) -> AsyncIterator[ExampleModel]:
        yield ExampleModel(value="example")

    def no_typehint(self):
        pass

    def explicit_starlette_response(self) -> JSONResponse:
        return JSONResponse(content=dict(value="example"))

    # Regular payload return functions
    assert extract_response_model_from_signature(
        none_typehint,
    ) == (None, ResponseModelType.SINGLE_RESPONSE)
    assert extract_response_model_from_signature(
        example_typehint,
    ) == (ExampleModel, ResponseModelType.SINGLE_RESPONSE)
    assert extract_response_model_from_signature(
        example_async_typehint,
    ) == (ExampleModel, ResponseModelType.SINGLE_RESPONSE)

    # Iterator return functions
    assert extract_response_model_from_signature(
        example_iterator_typehint,
    ) == (ExampleModel, ResponseModelType.ITERATOR_RESPONSE)

    assert extract_response_model_from_signature(
        example_async_iterator_typehint,
    ) == (ExampleModel, ResponseModelType.ITERATOR_RESPONSE)

    assert extract_response_model_from_signature(
        explicit_starlette_response,
    ) == (None, ResponseModelType.SINGLE_RESPONSE)

    # Deprecated but test until we move support for explicit_response
    with pytest.warns(DeprecationWarning):
        assert (
            extract_response_model_from_signature(no_typehint, ExampleModel)
            == (
                ExampleModel,
                ResponseModelType.SINGLE_RESPONSE,
            )
            == (ExampleModel, ResponseModelType.SINGLE_RESPONSE)
        )

    with pytest.warns(DeprecationWarning):
        assert extract_response_model_from_signature(
            no_typehint,
        ) == (None, ResponseModelType.SINGLE_RESPONSE)



from datetime import datetime
from typing import Any

from pydantic import BaseModel

from mountaineer.__tests__.client_builder.interface_builders.common import (
    create_exception_wrapper,
    create_field_wrapper,
    create_model_wrapper,
)
from mountaineer.client_builder.interface_builders.exception import ExceptionInterface
from mountaineer.client_builder.types import DictOf, ListOf
from mountaineer.exceptions import APIException


class TestBasicGeneration:
    def test_simple_exception(self):
        wrapper = create_exception_wrapper(
            APIException,
            "SimpleException",
            400,
            [create_field_wrapper("message", str), create_field_wrapper("code", int)],
        )

        interface = ExceptionInterface.from_exception(wrapper)
        ts_code = interface.to_js()

        assert "export interface SimpleException" in ts_code
        assert "message: string" in ts_code
        assert "code: number" in ts_code

    def test_optional_fields(self):
        wrapper = create_exception_wrapper(
            APIException,
            "OptionalFieldsException",
            400,
            [
                create_field_wrapper("message", str),
                create_field_wrapper("details", str, required=False),
                create_field_wrapper("timestamp", datetime, required=False),
            ],
        )

        interface = ExceptionInterface.from_exception(wrapper)
        ts_code = interface.to_js()

        assert "message: string" in ts_code
        assert "details?: string" in ts_code
        assert "timestamp?: string" in ts_code  # datetime converts to string

    def test_complex_fields(self):
        class ValidationError(BaseModel):
            field: str
            message: str

        class ErrorDetail(BaseModel):
            code: str
            description: str

        wrapper = create_exception_wrapper(
            APIException,
            "ComplexException",
            500,
            [
                create_field_wrapper("data", DictOf(str, Any)),
                create_field_wrapper(
                    "errors",
                    ListOf(create_model_wrapper(ValidationError, "ValidationError")),
                ),
                create_field_wrapper(
                    "details", create_model_wrapper(ErrorDetail, "ErrorDetail")
                ),
            ],
        )

        interface = ExceptionInterface.from_exception(wrapper)
        ts_code = interface.to_js()

        assert "data: Record<string, any>" in ts_code
        assert "errors: Array<ValidationError>" in ts_code
        assert "details: ErrorDetail" in ts_code



from mountaineer.app import AppController
from mountaineer.client_compiler.postcss import PostCSSBundler
from mountaineer.render import LinkAttribute, Metadata

{% if create_stub_files %}
from {{project_name}}.controllers.detail import DetailController
from {{project_name}}.controllers.home import HomeController
{% endif %}
from {{project_name}}.config import AppConfig

controller = AppController(
    config=AppConfig(), # type: ignore
    {% if use_tailwind %}
    global_metadata=Metadata(
        links=[LinkAttribute(rel="stylesheet", href="/static/app_main.css")]
    ),
    custom_builders=[
        PostCSSBundler(),
    ],
    {% endif %}
)

{% if create_stub_files %}
controller.register(HomeController())
controller.register(DetailController())
{% endif %}



from dataclasses import dataclass
from os.path import commonpath
from pathlib import Path
from re import finditer as re_finditer, sub
from time import monotonic_ns

from pydantic import BaseModel

from mountaineer import mountaineer as mountaineer_rs  # type: ignore
from mountaineer.logging import LOGGER


@dataclass
class ValueMask:
    mask: int
    right_padding: int


class SourceMapSchema(BaseModel):
    version: int
    sources: list[str]
    names: list[str]
    mappings: str
    sourcesContent: list[str] | None = None
    sourceRoot: str | None = None
    file: str | None = None


class SourceMapParser:
    """
    Parse sourcemaps according to the official specification:
    https://sourcemaps.info/spec.html
    """

    def __init__(
        self,
        path: str | Path | None = None,
        script: str | None = None,
    ):
        self.path = Path(path) if path else None
        self.script = script

        self.source_map: SourceMapSchema | None = None
        self._common_prefix_cache: dict[frozenset[str], str | None] = {}

        # { (line, column) : MapMetadata }
        self.parsed_mappings: (
            dict[tuple[int, int], mountaineer_rs.MapMetadata] | None
        ) = None

    def find_common_prefix(self, paths: list[str]) -> str | None:
        """
        Find the common prefix among all non-anonymous paths.
        Caches results based on the set of input paths.
        """
        # Convert paths to a frozenset for cache key
        paths_set = frozenset(paths)

        # Check cache first
        if paths_set in self._common_prefix_cache:
            return self._common_prefix_cache[paths_set]

        # Filter out anonymous paths and empty paths
        valid_paths = [
            p for p in paths if p and not p.startswith("<") and not p.endswith(">")
        ]

        if not valid_paths:
            result = None
        else:
            try:
                common = commonpath(valid_paths)
                result = common if common != "/" else None
            except ValueError:
                result = None

        # Cache the result
        self._common_prefix_cache[paths_set] = result
        return result

    def parse(self):
        """
        Parse the source map file and build up the internal mappings.
        Common prefix calculation is deferred until needed.
        """
        # If we've already parsed this file, don't do it again
        if self.parsed_mappings is not None:
            return

        text = Path(self.path).read_text() if self.path else self.script
        if not text:
            raise ValueError("No source map found")

        start_parse = monotonic_ns()
        self.source_map = SourceMapSchema.model_validate_json(text)
        LOGGER.debug(
            f"Parsed source map in {(monotonic_ns() - start_parse) / 1e9:.2f}s"
        )

        start_parse = monotonic_ns()
        self.parsed_mappings = mountaineer_rs.parse_source_map_mappings(
            self.source_map.mappings
        )
        LOGGER.debug(f"Parsed mappings in {(monotonic_ns() - start_parse) / 1e9:.2f}s")

    def get_original_location(self, line: int, column: int):
        """
        For a compiled line and column, return the original line and column where they appeared
        in the pre-built file.
        """
        if self.parsed_mappings is None:
            raise ValueError("SourceMapParser has not been parsed yet")

        return self.parsed_mappings.get((line, column))

    def map_exception(self, exception: str) -> str:
        """
        Given a JS stack exception, try to map it to the original files and line numbers

        :param exception: The exception string to map

        :return: The exception string with the original file and line numbers. Note that some
            exception stack traces may not be mappable, and will be left as-is.
        """
        if self.source_map is None or self.parsed_mappings is None:
            raise ValueError("SourceMapParser has not been parsed yet")

        # Build up the replacements all at once, since the matched indexes will be tied
        # to the original string
        text_replacements: dict[tuple[int, int], str] = {}
        relevant_sources = set()

        # First pass: collect all relevant source indices
        for match in re_finditer(
            r"\(([<>A-Za-z0-9/_.()]+?):(\d+?):(\d+?)\)", exception
        ):
            original_match = self.get_original_location(
                int(match.group(2)), int(match.group(3))
            )
            if original_match and original_match.source_index is not None:
                source = self.source_map.sources[original_match.source_index]
                relevant_sources.add(source)

        # Only calculate common prefix for sources that are actually used
        common_prefix = self.find_common_prefix(list(relevant_sources))

        # Second pass: build replacements
        for match in re_finditer(
            r"\(([<>A-Za-z0-9/_.()]+?):(\d+?):(\d+?)\)", exception
        ):
            original_match = self.get_original_location(
                int(match.group(2)), int(match.group(3))
            )

            if original_match and original_match.source_index is not None:
                source = self.source_map.sources[original_match.source_index]
                text_replacements[match.span(1)] = self._convert_relative_path(
                    source, common_prefix
                )
                text_replacements[match.span(2)] = str(original_match.source_line)
                text_replacements[match.span(3)] = str(original_match.source_column)

        # Sort in reverse order based on their start index to ensure that modifying parts of the string
        # doesn't affect the positions of parts that haven't been modified yet
        sorted_replacements = sorted(
            text_replacements.items(), key=lambda x: x[0][0], reverse=True
        )

        for (start, end), replacement in sorted_replacements:
            exception = exception[:start] + replacement + exception[end:]

        return exception

    def _convert_relative_path(self, path: str, common_prefix: str | None) -> str:
        """
        Convert a path by stripping common prefix and cleaning up.
        - Keeps anonymous paths (<anonymous>) unchanged
        - Strips common prefix from regular paths
        - Removes leading "../" sequences
        """
        # Don't modify anonymous paths
        if path.startswith("<") and path.endswith(">"):
            return path

        # Strip the common prefix if it exists
        if common_prefix and path.startswith(common_prefix):
            path = path[len(common_prefix) :].lstrip("/")

        return path


def get_cleaned_js_contents(contents: str):
    """
    Strip all single or multiline comments, since these can be dynamically generated
    metadata and can change without the underlying logic changing.
    """
    return mountaineer_rs.strip_js_comments(contents).strip()


def update_source_map_path(contents: str, new_path: str):
    """
    Updates the source map path to the new path, since the path is dynamic.
    """
    return sub(r"sourceMappingURL=(.*?).map", f"sourceMappingURL={new_path}", contents)



import asyncio
from multiprocessing import Queue
from threading import Thread

from fastapi import FastAPI, WebSocket
from starlette.websockets import WebSocketDisconnect

from mountaineer.development.uvicorn import UvicornThread
from mountaineer.io import get_free_port
from mountaineer.logging import LOGGER


class WatcherWebservice:
    """
    A simple webserver to notify frontends about updated builds via WebSockets.

    The WatcherWebservice provides a multiprocessing safe queue that's
    accessible as `notification_queue`. Each time that a process
    wants to update the frontend, it can push a message into the queue.

    This service is a critical component in the hot-reloading architecture,
    enabling backend changes to trigger frontend refreshes without manual intervention.
    It runs as a lightweight FastAPI application in a separate thread.

    """

    def __init__(self, webservice_host: str, webservice_port: int | None = None):
        """
        Initialize a new WatcherWebservice for notifying frontends about build updates.

        :param webservice_host: The host address to bind the webservice to (e.g. '127.0.0.1')
        :param webservice_port: Optional port number to use. If not provided, a free port will be allocated.
        """
        self.app = self.build_app()
        self.websockets: list[WebSocket] = []
        self.host = webservice_host
        self.port = webservice_port or get_free_port()
        self.notification_queue: Queue[bool | None] = Queue()

        self.webservice_thread: UvicornThread | None = None
        self.monitor_build_thread: Thread | None = None

        self.has_started = False

    def build_app(self):
        """
        Construct the FastAPI application with necessary routes and WebSocket endpoints.

        Creates a simple API with:
        - A root endpoint returning basic service information
        - A WebSocket endpoint at "/build-events" for real-time build notifications

        :return: Configured FastAPI application instance
        """
        app = FastAPI()

        @app.get("/")
        def home():
            return {"message": "This is the mountaineer hot-reload server."}

        @app.websocket("/build-events")
        async def build_updated(websocket: WebSocket):
            await websocket.accept()
            self.websockets.append(websocket)

            while True:
                # Keep the connection open until the client side disconnects
                try:
                    data = await websocket.receive_text()
                except WebSocketDisconnect:
                    # Expected exception, don't log an error
                    self.websockets.remove(websocket)
                    break
                await websocket.send_text(f"echo: {data}")

        return app

    async def broadcast_listeners(self):
        """
        Send a notification to all connected WebSocket clients.

        This method is called when a new build is available, signaling
        all connected frontends to refresh.
        """
        for ws in self.websockets:
            await ws.send_text("Build updated")
        LOGGER.info("Broadcasted build update to %d listeners", len(self.websockets))

    def monitor_builds(self):
        """
        Monitor the notification queue and broadcast updates to connected clients.

        This method runs in a separate thread, blocking on the notification queue
        and calling broadcast_listeners when a notification is received. The thread
        terminates when None is pushed to the queue.
        """
        while True:
            next_obj = self.notification_queue.get()
            if next_obj is None:
                break

            # Run in another thread's context
            asyncio.run(self.broadcast_listeners())

    async def start(self):
        """
        Start the WatcherWebservice by launching the necessary threads.

        Initializes and starts:
        - A UvicornThread to serve the FastAPI application
        - A monitor thread to watch for build notifications

        :raises Exception: If the service has already been started
        """
        if self.has_started:
            raise Exception("WatcherWebservice has already started")

        # UvicornThreads are daemon threads by default
        self.webservice_thread = UvicornThread(
            name="Hot reload server",
            emoticon="🦉",
            app=self.app,
            host=self.host,
            port=self.port,
            log_level="warning",
            use_logs=False,
        )

        LOGGER.debug("Starting WatcherWebservice on port %d", self.port)
        await self.webservice_thread.astart()

        LOGGER.debug("Starting monitor build thread")
        self.monitor_build_thread = Thread(target=self.monitor_builds, daemon=True)
        self.monitor_build_thread.start()

        LOGGER.debug("Finished WatcherWebservice constructor")
        self.has_started = True

    async def stop(self, wait_for_completion: int = 1) -> bool:
        """
        Attempts to stop the separate WatcherWebservice threads.

        Sends termination signals to the Uvicorn server and monitor threads,
        then waits for them to complete within the specified timeout period.

        :param wait_for_completion: Seconds to wait for threads to terminate (default: 1)
        :return: True if all threads terminated successfully, False otherwise
        """
        success: bool = True
        if self.webservice_thread is not None:
            await self.webservice_thread.astop()
            self.webservice_thread.join(wait_for_completion)
        if self.monitor_build_thread is not None:
            self.notification_queue.put(None)
            self.monitor_build_thread.join(wait_for_completion)

        if (self.webservice_thread and self.webservice_thread.is_alive()) or (
            self.monitor_build_thread and self.monitor_build_thread.is_alive()
        ):
            success = False
            LOGGER.info(
                f"WatcherWebservice still has outstanding threads: {self.webservice_thread} {self.monitor_build_thread}"
            )
        else:
            LOGGER.info("WatcherWebservice has fully stopped")

        return success



from abc import ABC, abstractmethod
from os import environ
from pathlib import Path
from subprocess import Popen

import toml
from click import secho


class EnvironmentBase(ABC):
    """
    Base class for backend environments where we'll install the python
    dependencies in the remote project, so we avoid a system wide install.

    """

    def __init__(self):
        self.global_env = {
            "HOME": environ.get("HOME", ""),
            "MOUNTAINEER_LOG_LEVEL": environ.get("MOUNTAINEER_LOG_LEVEL", "WARNING"),
        }

    @abstractmethod
    def has_provider(self) -> bool:
        """
        Determines whether the environment can be executed locally
        by a given provider

        """
        pass

    @abstractmethod
    def install_project(self, project_path: Path) -> None:
        """
        Perform the installation of the python dependencies into
        the current project.

        """
        pass

    def insert_wheel(
        self, package_name: str, wheel_path: Path, project_path: Path
    ) -> None:
        """
        Update the project's dependency configuration to use a local wheel file instead
        of pulling from PyPI.

        :param package_name: Name of the package to replace
        :param wheel_path: Path to the wheel file to use
        :param project_path: Path to the project root
        """
        pyproject_path = project_path / "pyproject.toml"
        if not pyproject_path.exists():
            raise ValueError("pyproject.toml not found")

        # Read the current pyproject.toml
        with open(pyproject_path) as f:
            pyproject = toml.load(f)

        # Update the dependency to point to the wheel file
        if "project" not in pyproject:
            pyproject["project"] = {}
        if "dependencies" not in pyproject["project"]:
            pyproject["project"]["dependencies"] = []

        # Remove any existing dependency for this package
        pyproject["project"]["dependencies"] = [
            dep
            for dep in pyproject["project"]["dependencies"]
            if not (isinstance(dep, str) and dep.startswith(package_name))
        ]

        # Add the wheel file as a dependency using PEP 508 format
        # Format: package_name @ file:///absolute/path/to/wheel
        wheel_path_str = wheel_path.resolve().as_uri()
        pyproject["project"]["dependencies"].append(
            f"{package_name} @ {wheel_path_str}"
        )

        secho(
            f"Updated pyproject.toml to use local wheel: {pyproject['project']['dependencies']}",
            fg="blue",
        )

        # Write the updated pyproject.toml
        with open(pyproject_path, "w") as f:
            toml.dump(pyproject, f)

    @abstractmethod
    def run_command(self, command: list[str], path: Path) -> Popen:
        """
        Run a command in the environment

        """
        pass

    @abstractmethod
    def get_env_path(self, project_path: Path) -> str:
        """
        Get the environment path. Only valid after the installation
        process has completed. Otherwise will raise an error.

        """
        pass



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from mountaineer.client_compiler.postcss import PostCSSBundler
from mountaineer.paths import ManagedViewPath


@pytest.mark.parametrize(
    "view_path, expected_output_name",
    [
        (
            "root_style.css",
            "root_style.css",
        ),
        (
            "root_style.scss",
            "root_style.css",
        ),
        (
            "home/style.css",
            "home_style.css",
        ),
        (
            "home/nested/other/style.css",
            "home_nested_other_style.css",
        ),
    ],
)
def test_get_style_output_name(view_path: str, expected_output_name: str):
    view_root = ManagedViewPath.from_view_root("fake-root")

    bundler = PostCSSBundler()
    assert bundler.get_style_output_name(view_root / view_path) == expected_output_name


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "css_path_str",
    [
        "styles.css",  # Standard case
        "components/nested/styles.css",  # Nested case
        "styles.scss",  # File with different extension
    ],
)
async def test_process_css_uses_absolute_paths(css_path_str: str):
    # Setup test data and paths
    package_root = ManagedViewPath.from_view_root("/fake/root")
    view_root = ManagedViewPath.from_view_root("/fake/root/views")
    css_path = view_root / css_path_str

    # Setup bundler with metadata
    bundler = PostCSSBundler()
    bundler.metadata = MagicMock()
    bundler.metadata.package_root_link = package_root

    # Create cli_path for the mock
    cli_path = Path("/fake/root/views/node_modules/.bin/postcss")

    # Mock the process result
    mock_process = AsyncMock()
    mock_process.communicate.return_value = (b"", b"")
    mock_process.returncode = 0

    with (
        TemporaryDirectory() as temp_dir_name,
        patch.object(bundler, "postcss_is_installed", return_value=(True, cli_path)),
        patch("asyncio.create_subprocess_exec", return_value=mock_process) as mock_exec,
        patch("tempfile.TemporaryDirectory") as mock_temp_dir,
    ):
        temp_dir_path = Path(temp_dir_name)
        mock_temp_dir.return_value.__enter__.return_value = str(temp_dir_path)

        # Mock output file
        with patch.object(Path, "read_text", return_value="processed CSS"):
            # Call the method
            result = await bundler.process_css(css_path)

            # Verify the result
            assert result == "processed CSS"

            # Verify subprocess was called with absolute paths
            mock_exec.assert_called_once()
            args, kwargs = mock_exec.call_args

            # Check command has absolute paths
            assert str(cli_path.absolute()) == args[0]
            assert str(css_path.absolute()) == args[1]
            assert "-o" == args[2]

            # Check that env params are set
            assert kwargs.get("cwd") == css_path.get_root_link().absolute()

            env = kwargs.get("env", {})
            assert "NODE_PATH" in env
            assert str(package_root.absolute() / "node_modules") == env["NODE_PATH"]



import os
import subprocess
from shutil import copy2, copytree

import pytest

from mountaineer.logging import LOGGER
from mountaineer.static import get_static_path


def run_command(command: str, cwd: str) -> tuple[str, str]:
    try:
        # Use subprocess.Popen to capture output in real-time
        process = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            cwd=cwd,
            env=dict(os.environ, FORCE_COLOR="true"),  # Force color output
        )

        # Capture both stdout and stderr
        stdout, stderr = process.communicate()

        if process.returncode != 0:
            raise subprocess.CalledProcessError(
                process.returncode, command, stdout, stderr
            )

        return stdout, stderr
    except subprocess.CalledProcessError as e:
        LOGGER.error(f"Error executing command: {command}")
        LOGGER.error(f"Error output:\n{e.stderr}")
        raise


@pytest.fixture(scope="session")
def setup_test_environment(tmp_path_factory: pytest.TempPathFactory):
    temp_dir = tmp_path_factory.mktemp("test_environment")
    harness_path = get_static_path(".")

    # Copy all files from harness directory to temp directory
    for item in os.listdir(harness_path):
        s = os.path.join(harness_path, item)
        d = os.path.join(temp_dir, item)
        if os.path.isdir(s):
            copytree(s, d)
        else:
            copy2(s, d)

    # The harness with package.json lives within one nested directory
    temp_dir = temp_dir / "harness"

    try:
        run_command("npm install", cwd=str(temp_dir))
        yield temp_dir
    finally:
        LOGGER.info(f"Test environment cleaned up at {temp_dir}")


def test_static_frontend(setup_test_environment: str):
    try:
        # Jest should return a non-zero exit code if tests fail, so if it came
        # back normally we can assume the tests passed
        stdout, stderr = run_command("npm run test", cwd=setup_test_environment)
        print(f"Jest stdout:\n{stdout}")  # noqa: T201
        print(f"Jest stderr:\n{stderr}")  # noqa: T201
    except subprocess.CalledProcessError as e:
        LOGGER.error(f"Jest tests failed. Error output:\n{e.stderr}")
        LOGGER.error(f"Jest stdout output:\n{e.stdout}")
        raise



import json
import logging
import os
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from io import StringIO

import pytest

from mountaineer.logging import debug_log_artifact, reset_artifact_dir, setup_logger


@contextmanager
def capture_logger_output(logger_func, *args, **kwargs):
    new_stdout, new_stderr = StringIO(), StringIO()
    with redirect_stdout(new_stdout), redirect_stderr(new_stderr):
        logger = logger_func(*args, **kwargs)
        yield logger, new_stdout, new_stderr


def test_setup_logger_name_and_level():
    logger_name = "test_logger"
    log_level = logging.DEBUG
    with capture_logger_output(setup_logger, logger_name, log_level) as (logger, _, _):
        assert logger.name == logger_name
        assert logger.level == log_level


def test_log_message_output():
    log_message = "Test message"
    logger_name = "output_logger"
    with capture_logger_output(setup_logger, logger_name) as (logger, stdout, _):
        logger.info(log_message)
        output = stdout.getvalue()
    # Since actual output includes ANSI escape codes for colors, we validate the presence of the message and level
    assert log_message in output
    assert "INFO" in output


def test_timestamp():
    logger_name = "format_logger"
    with capture_logger_output(setup_logger, logger_name) as (logger, stdout, _):
        logger.warning("Warning test")
        output = stdout.getvalue()
    log_output = json.loads(output.strip())
    assert "timestamp" in log_output


def test_exception_logging():
    logger_name = "exception_logger"
    with capture_logger_output(setup_logger, logger_name) as (logger, stdout, _):
        try:
            raise ValueError("Test exception")
        except ValueError:
            logger.exception("This is an exception")
        output = stdout.getvalue()
        log_output = json.loads(output.strip())
        assert "exception" in log_output
        assert "Test exception" in log_output["exception"]


@contextmanager
def modify_log_level(log_level: str):
    current_log_level = os.environ.get("MOUNTAINEER_LOG_LEVEL")
    os.environ["MOUNTAINEER_LOG_LEVEL"] = log_level
    try:
        yield current_log_level
    finally:
        if current_log_level is not None:
            os.environ["MOUNTAINEER_LOG_LEVEL"] = current_log_level
        else:
            del os.environ["MOUNTAINEER_LOG_LEVEL"]


@pytest.mark.parametrize(
    "log_level,should_create_file",
    [
        ("DEBUG", True),
        ("INFO", False),
        ("WARNING", False),
        ("ERROR", False),
    ],
)
def test_debug_log_artifact(log_level: str, should_create_file: bool):
    # Set up test environment
    test_content = "test content"
    test_prefix = "test"
    test_ext = "txt"

    # Reset any global state every run
    reset_artifact_dir()

    with modify_log_level(log_level):
        path = debug_log_artifact(test_prefix, test_ext, test_content)

    # Check the results
    if should_create_file:
        assert path is not None
        tmp_path = path.parent
        files = list(tmp_path.glob(f"{test_prefix}-*.{test_ext}"))

        assert len(files) == 1
        with open(files[0], "r") as f:
            assert f.read() == test_content
    else:
        assert path is None



import pytest
from pydantic import Field, ValidationError

from mountaineer.exceptions import APIException


def test_exceptions():
    class Test402(APIException):
        status_code = 402
        detail = "Test 402"
        custom_value: str

    exception = Test402(custom_value="custom")

    # Make sure that we pass along the metadata to the exception superclass
    assert exception.status_code == 402
    assert exception.detail == "Test 402"
    assert exception.headers == {}
    # assert exception.custom_value == "custom"

    # Test its internal value
    assert exception.internal_model.status_code == 402
    assert exception.internal_model.detail == "Test 402"

    # We need to ignore typing here because custom_value is dynamically
    # added to the pydantic schema via create_model
    # It won't typehint correctly but it will be available at runtime
    # which is all we care about
    assert exception.internal_model.custom_value == "custom"  # type: ignore


def test_internal_model_constructor():
    """
    Test the translation of the APIException to an internal Pydantic model with
    the correct internal types.

    """

    class Test402(APIException):
        status_code = 402
        detail = "Test 402"
        custom_value: str

    assert not Test402.InternalModel.model_fields["status_code"].is_required()
    assert not Test402.InternalModel.model_fields["detail"].is_required()
    assert Test402.InternalModel.model_fields["custom_value"].is_required()


def test_exceptions_validate_values():
    class Test402(APIException):
        status_code = 402
        detail = "Test 402"
        custom_value: int = Field(lt=10)

    with pytest.raises(ValidationError):
        Test402(custom_value=20)



import subprocess
from os import environ
from pathlib import Path

from click import secho

from create_mountaineer_app.environments.base import EnvironmentBase


class UvEnvironment(EnvironmentBase):
    """
    Use uv as the Python package installer and environment manager.
    uv is a modern, fast package installer and resolver written in Rust.
    """

    def __init__(self, venv_name: str = ".venv"):
        super().__init__()
        self.venv_name = venv_name
        self.limited_scope_env = {
            "PATH": environ["PATH"],
            **self.global_env,
        }

    def has_provider(self) -> bool:
        try:
            # Attempt to get the version of uv to check if it's installed
            subprocess.run(
                ["uv", "--version"],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return True
        except subprocess.CalledProcessError:
            # uv is installed but there might be a problem with it
            return True
        except FileNotFoundError:
            # uv is not installed
            return False

    def install_project(self, project_path: Path) -> None:
        # Create a virtual environment using uv
        venv_path = project_path / self.venv_name
        subprocess.run(
            ["uv", "venv", str(venv_path)],
            check=True,
            cwd=str(project_path),
            env=self.limited_scope_env,
        )

        secho(f"Virtual environment created at: {venv_path}", fg="green")

        # Install packages using uv pip
        subprocess.run(
            ["uv", "pip", "install", "-e", "."],
            check=True,
            cwd=project_path,
            env={
                "VIRTUAL_ENV": str(venv_path),
                "PATH": f"{venv_path}/bin:{environ['PATH']}",
                **self.global_env,
            },
        )

        secho("Packages installed.", fg="green")

    def install_provider(self) -> None:
        """
        Install uv using the official installation method:
        > curl -LsSf https://astral.sh/uv/install.sh | sh

        Raises an exception if the installation fails.
        """
        subprocess.run(
            ["curl", "-LsSf", "https://astral.sh/uv/install.sh", "|", "sh"],
            check=True,
            shell=True,  # Required for pipe operation
        )
        secho("uv installed successfully.", fg="green")

    def run_command(self, command: list[str], path: Path):
        venv_path = path / self.venv_name
        if not venv_path.exists():
            raise ValueError(f"Virtual environment not found at: {venv_path}")

        return subprocess.Popen(
            command,
            cwd=path,
            env={
                "VIRTUAL_ENV": str(venv_path),
                "PATH": f"{venv_path}/bin:{environ['PATH']}",
                **self.global_env,
            },
        )

    def get_env_path(self, project_path: Path) -> str:
        venv_path = project_path / self.venv_name
        if not venv_path.exists():
            raise ValueError(f"Virtual environment not found at: {venv_path}")

        return str(venv_path)
